# import libs
import logging
from typing import List, Any, Dict, Optional
import pycuc

# NOTE: set up logger
logger = logging.getLogger(__name__)


def normalize_unit(
    data: List[Any],
    to_unit: str,
    valid_from: Optional[List[str]] = None,
) -> Dict:
    '''
    Normalizes a list of units to the specified unit.

    Parameters
    ----------
    data : List[Any]
        List of data with units.
    to_unit : str
        Unit to normalize to.
    valid_from : List[str], optional
        List of valid units to convert from, by default None.

    Returns
    -------
    Dict
        A dictionary containing the normalized data and target unit.
    '''
    try:
        # NOTE: check inputs
        to = to_unit.strip()

        # >> check
        if not to:
            logger.warning("Target unit is empty!")
            return {}

        # >> check
        if not data or not isinstance(data, list):
            logger.warning("Input data is empty or not a list!")
            return {}

        # NOTE: normalize units
        normalized_data: List[float] = []

        # >> loop over data
        for item in data:
            # >> get value and unit
            if hasattr(item, 'value') and hasattr(item, 'unit'):
                value = item.value
                from_unit = item.unit
            else:
                logger.warning(
                    f"Item {item} does not have 'value' and 'unit' attributes!")
                continue

            # >> check valid_from
            if (
                valid_from and
                from_unit not in valid_from
            ):
                logger.warning(f"Unit '{from_unit}' not in valid_from list!")
                continue

            # >> convert unit
            try:
                converted_value = pycuc.convert_from_to(
                    value=value,
                    from_unit=from_unit,
                    to_unit=to
                )
                normalized_data.append(converted_value)
            except Exception as e:
                logger.error(
                    f"Conversion failed for value {value} from '{from_unit}' to '{to}': ", e)
                continue

        return {
            "data": normalized_data,
            "to": to,
        }
    except Exception as e:
        logger.error("Unit normalization failed!, ", e)
        return {}



# import libs
from __future__ import annotations
import logging
import pandas as pd
import numpy as np
import math
from typing import Optional, Tuple, Dict, Any, List
from scipy.optimize import least_squares
from pathlib import Path
import pycuc
# local

# NOTE: set up logger
logger = logging.getLogger(__name__)


class Antoine:
    """
    Pure component vapor pressure model using Antoine equation as follows:

    - VaPr(T) = 10^(A - B/(T + C))
    - VaPr(T) = exp(A - B/(T + C))
    """

    def __init__(self):
        """
        Initialize with vapor pressure and derivative functions.
        """
        pass

    @staticmethod
    def fit_antoine(
        T_data: np.ndarray,
        P_data: np.ndarray,
        *,
        base: str = "log10",
        T_unit: str = "K",
        p_unit: str = "Pa",
        fit_in_log_space: bool = True,
        weights: Optional[np.ndarray] = None,
        x0: Optional[Tuple[float, float, float]] = None,
        bounds: Optional[
            Tuple[
                Tuple[float, float, float],
                Tuple[float, float, float]
            ]
        ] = None,
        max_nfev: int = 5000,
        validate: bool = True,
        min_margin_kelvin: float = 1.0,
        # robust options
        loss: str = "linear",
        f_scale: Optional[float] = None,
    ) -> Dict[str, Any]:
        """
        Fit Antoine coefficients (A,B,C) to experimental VaPr(T) data.

        - Supports T in K or °C via T_unit.
        - Supports P in Pa or bar via p_unit.
        - Recommended: fit_in_log_space=True.
        - Robust fitting supported via loss + f_scale.

        Parameters
        ----------
        T_data : np.ndarray
            Array of temperature data points.
        P_data : np.ndarray
            Array of vapor pressure data points.
        base : str, optional
            Logarithm base for Antoine equation: 'log10' or 'ln' (default 'log10').
        T_unit : str, optional
            Unit of temperature data: 'K' or 'C' (default 'K').
        p_unit : str, optional
            Unit of pressure data: 'Pa' or 'bar' (default 'Pa').
        fit_in_log_space : bool, optional
            If True, fit in log space; else fit in pressure space (default True).
        weights : Optional[np.ndarray], optional
            Optional weights for each data point (default None = equal weights).
        x0 : Optional[Tuple[float, float, float]], optional
            Initial guess for (A, B, C) (default None = auto guess).
        bounds : Optional[Tuple[Tuple[float, float, float], Tuple[float, float, float]]], optional
                Bounds for (A, B, C) (default None = wide bounds).
        max_nfev : int, optional
            Maximum number of function evaluations (default 5000).
        validate : bool, optional
            If True, perform fit validation checks (default True).
        min_margin_kelvin : float, optional
            Minimum margin for (T + C) to avoid denominator issues (default 1.0 K).
        loss : str, optional
            Loss function for robust fitting: 'linear', 'soft_l1', 'huber', 'cauchy', 'arctan' (default 'linear').
        f_scale : Optional[float], optional
            Scaling parameter for robust loss (default None = auto).

        Returns
        -------
        Dict[str, Any]
            Returns a fit report dict with coefficients, metrics, warnings, and solver info.
        """
        # SECTION: Input validation and conversion
        # NOTE: Convert inputs to arrays
        T = np.asarray(T_data, dtype=float).ravel()
        P = np.asarray(P_data, dtype=float).ravel()

        # Check sizes
        if T.size != P.size or T.size < 3:
            logger.error(
                "T_data and P_data must have same length and at least 3 points.")
            return {}

        # NOTE: Normalize string inputs
        T_unit = T_unit.lower()
        base = base.lower()
        p_unit = p_unit.lower()
        loss = loss.lower()

        # >> Convert temperature to K
        if T_unit in ("k", "kelvin"):
            T_k = T
        elif T_unit in ("c", "degc", "celsius", "°c"):
            T_k = T + 273.15
        else:
            logger.error("T_unit must be 'K' or 'C'.")
            return {}

        # >> Convert pressure to Pa
        if p_unit == "pa":
            P_pa = P
        elif p_unit == "bar":
            P_pa = P * 1e5
        else:
            logger.error("p_unit must be 'Pa' or 'bar'.")
            return {}

        # >> Check pressures > 0
        if np.any(P_pa <= 0):
            logger.error("All pressures must be > 0 for Antoine fitting.")
            return {}

        # NOTE: Choose log target
        if base == "log10":
            y = np.log10(P_pa)
        elif base == "ln":
            y = np.log(P_pa)
        else:
            logger.error("base must be 'log10' or 'ln'.")
            return {}

        # NOTE: weights (sqrt form)
        if weights is None:
            w = np.ones_like(T_k)
        else:
            w_raw = np.asarray(weights, dtype=float).ravel()
            if w_raw.size != T_k.size:
                logger.error("weights must have same length as data.")
                return {}

            # Ensure non-negative weights
            w = np.sqrt(np.clip(w_raw, 0.0, np.inf))

        # SECTION: Define model and residuals
        def model_log(params: np.ndarray) -> np.ndarray:
            A_, B_, C_ = params
            return A_ - B_ / (T_k + C_)

        # NOTE: Define residuals
        if fit_in_log_space:
            def residuals(params: np.ndarray) -> np.ndarray:
                return w * (model_log(params) - y)
        else:
            if base == "log10":
                def Pmodel(params: np.ndarray) -> np.ndarray:
                    return 10.0 ** model_log(params)
            else:
                def Pmodel(params: np.ndarray) -> np.ndarray:
                    return np.exp(model_log(params))

            def residuals(params: np.ndarray) -> np.ndarray:
                return w * (Pmodel(params) - P_pa)

        # NOTE: Initial guess
        if x0 is None:
            C0 = -50.0
            if np.min(T_k + C0) <= min_margin_kelvin:
                # ! C0 too low, adjust
                C0 = -np.min(T_k) + 10.0

            # ! A0: mean logP
            A0 = float(np.mean(y))

            # ! B0: from rough slope
            m, _b = np.polyfit(1.0 / T_k, y, 1)

            Tmean = float(np.mean(T_k))
            ratio = (float(np.mean(T_k + C0)) / Tmean) ** 2
            B0 = float(abs(m) * ratio)

            # >> sanity check B0
            if not np.isfinite(B0) or B0 <= 1e-6:
                B0 = 2000.0

            # ! > assemble x0
            x0 = (A0, B0, C0)

        # >> Convert x0 to array
        x0_array = np.asarray(x0, dtype=float)

        # NOTE: Default bounds
        if bounds is None:
            bounds = ((-200.0, 1e-6, -1e4), (200.0, 1e7, 1e4))

        # NOTE: Auto f_scale for robust loss if not provided
        if f_scale is None:
            if loss != "linear":
                f_scale = 0.02 if fit_in_log_space else max(
                    1.0, float(np.median(P_pa) * 0.02))
            else:
                f_scale = 1.0

        # SECTION: Perform least squares fitting
        res = least_squares(
            residuals,
            x0=x0_array,
            bounds=bounds,
            max_nfev=max_nfev,
            loss=loss,
            f_scale=float(f_scale),
        )

        # NOTE: Extract fitted parameters
        A, B, C = map(float, res.x)

        # NOTE: Predictions and metrics
        y_hat = A - B / (T_k + C)
        if base == "log10":
            P_hat = 10.0 ** y_hat
        else:
            P_hat = np.exp(y_hat)

        log_res = y_hat - y

        # NOTE: Metrics
        rmse_log = float(np.sqrt(np.mean(log_res ** 2)))
        mae_log = float(np.mean(np.abs(log_res)))

        P_res = P_hat - P_pa
        rmse_P = float(np.sqrt(np.mean(P_res ** 2)))
        mae_P = float(np.mean(np.abs(P_res)))

        y_mean = float(np.mean(y))
        ss_res = float(np.sum((y - y_hat) ** 2))
        ss_tot = float(np.sum((y - y_mean) ** 2))
        r2_log = float(1.0 - ss_res / ss_tot) if ss_tot > 0 else float("nan")

        # NOTE: Covariance estimate (approx)
        cov = None
        try:
            J = res.jac
            dof = max(1, T_k.size - 3)
            s2 = 2.0 * float(res.cost) / dof
            cov = (s2 * np.linalg.inv(J.T @ J)).tolist()
        except Exception:
            cov = None

        # SECTION: Validation checks
        warnings: List[str] = []
        if validate:
            denom_min = float(np.min(T_k + C))
            if denom_min <= min_margin_kelvin:
                warnings.append(
                    f"Risky fit: min(T + C) = {denom_min:.6g} K (<= {min_margin_kelvin} K). "
                    "Denominator near zero can make the fit unstable."
                )

            idx = np.argsort(T_k)
            if np.any(np.diff(P_hat[idx]) <= 0):
                warnings.append(
                    "Non-physical trend: fitted VaPr(T) is not strictly increasing over the data range.")

            if B < 10.0:
                warnings.append(
                    "Unusually small B; data range may be too narrow or units may be inconsistent.")
            if B > 1e6:
                warnings.append(
                    "Very large B; check units and outliers in experimental data.")

        # SECTION: Return fit report
        return {
            "A": A, "B": B, "C": C,
            "base": base,
            "p_unit": "pa",
            "T_unit_internal": "K",
            "fit_in_log_space": bool(fit_in_log_space),
            "success": bool(res.success),
            "message": str(res.message),
            "cost": float(res.cost),
            "rmse_logP": rmse_log,
            "mae_logP": mae_log,
            "r2_logP": r2_log,
            "rmse_P": rmse_P,
            "mae_P": mae_P,
            "cov": cov,
            "warnings": warnings,
            "Tmin_K": float(np.min(T_k)),
            "Tmax_K": float(np.max(T_k)),
            # robust metadata
            "loss": loss,
            "f_scale": float(f_scale),
        }

    @staticmethod
    def _robust_weight(loss: str, z: float) -> float:
        """
        Approx IRLS-like weight for standardized residual z = r / f_scale.

        Parameters
        ----------
        loss : str
            Loss function: 'linear', 'soft_l1', 'huber', 'cauchy', 'arctan'.
        z : float
            Standardized residual.

        Returns
        -------
        float
            Robust weight for the given standardized residual.
        """
        a = abs(z)
        loss = loss.lower()

        if loss == "linear":
            return 1.0
        if loss == "soft_l1":
            return 1.0 / math.sqrt(1.0 + z * z)
        if loss == "huber":
            return 1.0 if a <= 1.0 else 1.0 / a
        if loss == "cauchy":
            return 1.0 / (1.0 + z * z)
        if loss == "arctan":
            return 1.0 / (1.0 + z * z)

        return 1.0

    @classmethod
    def outlier_report(
        cls,
        T_data: np.ndarray,
        P_data: np.ndarray,
        fit_report: Dict[str, Any],
        *,
        T_unit: str = "K",
        p_unit: str = "Pa",
        top_n: int = 10,
        residual_domain: str = "log",  # "log" or "P"
    ) -> List[Dict[str, Any]]:
        """
        Rank points by robust down-weighting and standardized residual magnitude.

        Parameters
        ----------
        T_data : np.ndarray
            Array of temperature data points.
        P_data : np.ndarray
            Array of vapor pressure data points.
        fit_report : Dict[str, Any]
            Fit report dict from fit_antoine().
        T_unit : str, optional
            Unit of temperature data: 'K' or 'C' (default 'K').
        p_unit : str, optional
            Unit of pressure data: 'Pa' or 'bar' (default 'Pa').
        top_n : int, optional
            Number of top outliers to report (default 10).
        residual_domain : str, optional
            Domain for residuals: 'log' or 'P' (default 'log').

        Returns
        -------
        List[Dict[str, Any]]
            Returns a list of dicts with:
            - index,
            - T_K,
            - P_input_Pa,
            - P_fit_Pa,
            - residual,
            - standardized_residual,
            - robust_weight
        """
        residual_domain = residual_domain.lower()

        T = np.asarray(T_data, dtype=float).ravel()
        P = np.asarray(P_data, dtype=float).ravel()

        if T_unit.lower() in ("c", "degc", "celsius", "°c"):
            T_k = T + 273.15
        else:
            T_k = T

        if p_unit.lower() == "bar":
            P_pa = P * 1e5
        else:
            P_pa = P

        A, B, C = float(fit_report["A"]), float(
            fit_report["B"]), float(fit_report["C"])
        base = str(fit_report.get("base", "log10")).lower()
        loss = str(fit_report.get("loss", "linear")).lower()
        f_scale = float(fit_report.get("f_scale", 1.0))

        y_hat = A - B / (T_k + C)
        if base == "log10":
            y = np.log10(P_pa)
            P_hat = 10.0 ** y_hat
        else:
            y = np.log(P_pa)
            P_hat = np.exp(y_hat)

        if residual_domain == "log":
            r = y_hat - y
        elif residual_domain == "p":
            r = P_hat - P_pa
        else:
            logger.error("residual_domain must be 'log' or 'P'.")
            return []

        # Standardized residuals
        z = r / f_scale if f_scale > 0 else np.zeros_like(r, dtype=float)
        w_rob = np.array(
            [cls._robust_weight(loss, zi) for zi in z],
            dtype=float
        )

        # Sort: smallest weight first, then largest standardized residual
        order = np.argsort(w_rob + 1e-12)  # primary by weight
        # refine within same-ish weights by |z|
        # secondary by abs(z), descending
        order = order[np.argsort(np.abs(z[order]))[::-1]]
        order = order[: min(top_n, order.size)]

        out: List[Dict[str, Any]] = []
        for i in order:
            out.append(
                {
                    "index": int(i),
                    "T_K": float(T_k[i]),
                    "P_input_Pa": float(P_pa[i]),
                    "P_fit_Pa": float(P_hat[i]),
                    "residual": float(r[i]),
                    "standardized_residual": float(z[i]),
                    "robust_weight": float(w_rob[i]),
                }
            )
        return out

    @staticmethod
    def load_experimental_data(
        experimental_data: str | Path,
        T_unit: str,
        P_unit: str,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Load experimental data from CSV file, then convert to arrays. The CSV file must contain 'Temperature' and 'Pressure' columns. Temperature and pressure units are specified via T_unit and P_unit, the defaults being 'K' and 'Pa' respectively.

        Parameters
        ----------
        experimental_data : str | Path
            Path to CSV file with 'Temperature' and 'Pressure' columns.
        T_unit : str
            Unit of temperature data: 'K' or 'C'.
        P_unit : str
            Unit of pressure data: 'Pa' or 'bar'.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Arrays of temperatures and pressures.
        """
        try:
            # SECTION: Load data
            df = pd.read_csv(experimental_data)

            # NOTE: Validate columns
            df_columns_lower = [col.strip().lower() for col in df.columns]

            # >> check columns
            if 'temperature' not in df_columns_lower or 'pressure' not in df_columns_lower:
                logger.error(
                    "CSV file must contain 'Temperature' and 'Pressure' columns.")
                return np.array([]), np.array([])
            # >> get actual column names
            temp_col = df.columns[df_columns_lower.index('temperature')]
            pres_col = df.columns[df_columns_lower.index('pressure')]

            # >> check data types
            if df[temp_col].dtype not in [np.float64, np.float32, np.int64, np.int32]:
                logger.error(
                    "'Temperature' column must contain numeric data.")
                return np.array([]), np.array([])

            if df[pres_col].dtype not in [np.float64, np.float32, np.int64, np.int32]:
                logger.error(
                    "'Pressure' column must contain numeric data.")
                return np.array([]), np.array([])

            # NOTE: >> to arrays
            temperatures = df[temp_col]
            # >> Unit conversions
            temperatures = [
                pycuc.convert_from_to(
                    float(T_val),
                    from_unit=T_unit,
                    to_unit='K'
                ) for T_val in temperatures
            ]

            pressures = df[pres_col]
            pressures = [
                pycuc.convert_from_to(
                    float(P_val),
                    from_unit=P_unit,
                    to_unit='Pa'
                ) for P_val in pressures
            ]

            # SECTION: to arrays
            pressures = np.array(pressures, dtype=float)
            temperatures = np.array(temperatures, dtype=float)

            return temperatures, pressures
        except Exception as e:
            logger.exception(
                f"Failed to load experimental data: {e}")
            return np.array([]), np.array([])



# import libs
from typing import Optional, Dict, TypedDict, Callable
from pydantic import BaseModel, Field, ConfigDict

# local
from .ref import GroupUnit

# SECTION: Joback Group Contributions


class JobackGroupContributions(BaseModel):
    """Joback group contribution identifiers and their corresponding contributions to thermodynamic properties."""
    model_config = ConfigDict(
        populate_by_name=True,
        arbitrary_types_allowed=True,
    )

    methyl: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Methyl group contribution.",
        alias="-CH3",
        json_schema_extra={
            "category": "non-ring increments", "group": "-CH3", "type": "normal", "id": "NR1", "Tc": 0.0141, "Pc": -0.0012, "Vc": 65.0, "Tb": 23.58, "Tf": -5.1, "EnFo_IG": -76.45, "GiEnFo_IG": -43.96, "a": 19.5, "b": -0.00808, "c": 0.000153, "d": -9.67e-08, "EnFus": 0.908, "EnVap": 2.373, "ηa": 548.29, "ηb": -1.719
        }
    )
    methylene: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Methylene group contribution.",
        alias="-CH2- @non-ring",
        json_schema_extra={
            "category": "non-ring increments", "group": "-CH2- @non-ring", "type": "normal", "id": "NR2", "Tc": 0.0189, "Pc": 0.0, "Vc": 56.0, "Tb": 22.88, "Tf": 11.27, "EnFo_IG": -20.64, "GiEnFo_IG": 8.42, "a": -0.909, "b": 0.095, "c": -5.44e-05, "d": 1.19e-08, "EnFus": 2.59, "EnVap": 2.226, "ηa": 94.16, "ηb": -0.199
        }
    )
    tertiary_CH: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Tertiary CH group contribution.",
        alias=">CH- @non-ring",
        json_schema_extra={
            "category": "non-ring increments", "group": ">CH- @non-ring", "type": "normal", "id": "NR3", "Tc": 0.0164, "Pc": 0.002, "Vc": 41.0, "Tb": 21.74, "Tf": 12.64, "EnFo_IG": 29.89, "GiEnFo_IG": 58.36, "a": -23.0, "b": 0.204, "c": -0.000265, "d": 1.2e-07, "EnFus": 0.749, "EnVap": 1.691, "ηa": -322.15, "ηb": 1.187
        }
    )
    quaternary_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Quaternary C group contribution.",
        alias=">C< @non-ring",
        json_schema_extra={
            "category": "non-ring increments", "group": ">C< @non-ring", "type": "normal", "id": "NR4", "Tc": 0.0067, "Pc": 0.0043, "Vc": 27.0, "Tb": 18.25, "Tf": 46.43, "EnFo_IG": 82.23, "GiEnFo_IG": 116.02, "a": -66.2, "b": 0.427, "c": -0.000641, "d": 3.01e-07, "EnFus": -1.46, "EnVap": 0.636, "ηa": -573.56, "ηb": 2.307
        }
    )
    vinyl_CH2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Vinyl CH2 group contribution.",
        alias="=CH2",
        json_schema_extra={
            "category": "non-ring increments", "group": "=CH2", "type": "normal", "id": "NR5", "Tc": 0.0113, "Pc": -0.0028, "Vc": 56.0, "Tb": 18.18, "Tf": -4.32, "EnFo_IG": -9.63, "GiEnFo_IG": 3.77, "a": 23.6, "b": -0.0381, "c": 0.000172, "d": -1.03e-07, "EnFus": -0.473, "EnVap": 1.724, "ηa": 495.01, "ηb": -1.539
        }
    )
    vinyl_CH: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Vinyl CH group contribution.",
        alias="=CH- @non-ring",
        json_schema_extra={
            "category": "non-ring increments", "group": "=CH- @non-ring", "type": "normal", "id": "NR6", "Tc": 0.0129, "Pc": -0.0006, "Vc": 46.0, "Tb": 24.96, "Tf": 8.73, "EnFo_IG": 37.97, "GiEnFo_IG": 48.53, "a": -8.0, "b": 0.105, "c": -9.63e-05, "d": 3.56e-08, "EnFus": 2.691, "EnVap": 2.205, "ηa": 82.28, "ηb": -0.242
        }
    )
    vinyl_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Vinyl C group contribution.",
        alias="=C< @non-ring",
        json_schema_extra={
            "category": "non-ring increments", "group": "=C< @non-ring", "type": "normal", "id": "NR7", "Tc": 0.0117, "Pc": 0.0011, "Vc": 38.0, "Tb": 24.14, "Tf": 11.14, "EnFo_IG": 83.99, "GiEnFo_IG": 92.36, "a": -28.1, "b": 0.208, "c": -0.000306, "d": 1.46e-07, "EnFus": 3.063, "EnVap": 2.138, "ηa": 0.0, "ηb": 0.0
        }
    )
    allene: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Allene group contribution.",
        alias="=C=",
        json_schema_extra={
            "category": "non-ring increments", "group": "=C=", "type": "normal", "id": "NR8", "Tc": 0.0026, "Pc": 0.0028, "Vc": 36.0, "Tb": 26.15, "Tf": 17.78, "EnFo_IG": 142.14, "GiEnFo_IG": 136.7, "a": 27.4, "b": -0.0557, "c": 0.000101, "d": -5.02e-08, "EnFus": 4.72, "EnVap": 2.661, "ηa": 0.0, "ηb": 0.0
        }
    )
    alkyne_CH: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Alkyne CH group contribution.",
        alias="#CH",
        json_schema_extra={
            "category": "non-ring increments", "group": "#CH", "type": "normal", "id": "NR9", "Tc": 0.0027, "Pc": -0.0008, "Vc": 46.0, "Tb": 9.2, "Tf": -11.18, "EnFo_IG": 79.3, "GiEnFo_IG": 77.71, "a": 24.5, "b": -0.0271, "c": 0.000111, "d": -6.78e-08, "EnFus": 2.322, "EnVap": 1.155, "ηa": 0.0, "ηb": 0.0
        }
    )
    alkyne_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Alkyne C group contribution.",
        alias="#C-",
        json_schema_extra={
            "category": "non-ring increments", "group": "#C-", "type": "normal", "id": "NR10", "Tc": 0.002, "Pc": 0.0016, "Vc": 37.0, "Tb": 27.38, "Tf": 64.32, "EnFo_IG": 115.51, "GiEnFo_IG": 109.82, "a": 7.87, "b": 0.0201, "c": -8.33e-06, "d": 1.39e-09, "EnFus": 4.151, "EnVap": 3.302, "ηa": 0.0, "ηb": 0.0
        }
    )
    methylene_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Methylene group contribution in ring.",
        alias="-CH2- @ring",
        json_schema_extra={
            "category": "ring increments", "group": "-CH2- @ring", "type": "normal", "id": "R1", "Tc": 0.01, "Pc": 0.0025, "Vc": 48.0, "Tb": 27.15, "Tf": 7.75, "EnFo_IG": -26.8, "GiEnFo_IG": -3.68, "a": -6.03, "b": 0.0854, "c": -8e-06, "d": -1.8e-08, "EnFus": 0.49, "EnVap": 2.398, "ηa": 307.53, "ηb": -0.798
        }
    )
    tertiary_CH_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Tertiary CH group contribution in ring.",
        alias=">CH- @ring",
        json_schema_extra={
            "category": "ring increments", "group": ">CH- @ring", "type": "normal", "id": "R2", "Tc": 0.0122, "Pc": 0.0004, "Vc": 38.0, "Tb": 21.78, "Tf": 19.88, "EnFo_IG": 8.67, "GiEnFo_IG": 40.99, "a": -20.5, "b": 0.162, "c": -0.00016, "d": 6.24e-08, "EnFus": 3.243, "EnVap": 1.942, "ηa": -394.29, "ηb": 1.251
        }
    )
    quaternary_C_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Quaternary C group contribution in ring.",
        alias=">C< @ring",
        json_schema_extra={
            "category": "ring increments", "group": ">C< @ring", "type": "normal", "id": "R3", "Tc": 0.0042, "Pc": 0.0061, "Vc": 27.0, "Tb": 21.32, "Tf": 60.15, "EnFo_IG": 79.72, "GiEnFo_IG": 87.88, "a": -90.9, "b": 0.557, "c": -0.0009, "d": 4.69e-07, "EnFus": -1.373, "EnVap": 0.644, "ηa": 0.0, "ηb": 0.0
        }
    )
    vinyl_CH_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Vinyl CH group contribution in ring.",
        alias="=CH- @ring",
        json_schema_extra={
            "category": "ring increments", "group": "=CH- @ring", "type": "normal", "id": "R4", "Tc": 0.0082, "Pc": 0.0011, "Vc": 41.0, "Tb": 26.73, "Tf": 8.13, "EnFo_IG": 2.09, "GiEnFo_IG": 11.3, "a": -2.14, "b": 0.0574, "c": -1.64e-06, "d": -1.59e-08, "EnFus": 1.101, "EnVap": 2.544, "ηa": 259.65, "ηb": -0.702
        }
    )
    vinyl_C_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Vinyl C group contribution in ring.",
        alias="=C< @ring",
        json_schema_extra={
            "category": "ring increments", "group": "=C< @ring", "type": "normal", "id": "R5", "Tc": 0.0143, "Pc": 0.0008, "Vc": 32.0, "Tb": 31.01, "Tf": 37.02, "EnFo_IG": 46.43, "GiEnFo_IG": 54.05, "a": -8.25, "b": 0.101, "c": -0.000142, "d": 6.78e-08, "EnFus": 2.394, "EnVap": 3.059, "ηa": -245.74, "ηb": 0.912
        }
    )
    fluorine: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Fluorine group contribution.",
        alias="-F",
        json_schema_extra={
            "category": "hologen increments", "group": "-F", "type": "normal", "id": "H1", "Tc": 0.0111, "Pc": -0.0057, "Vc": 27.0, "Tb": -0.03, "Tf": -15.78, "EnFo_IG": -251.92, "GiEnFo_IG": -247.19, "a": 26.5, "b": -0.0913, "c": 0.000191, "d": -1.03e-07, "EnFus": 1.398, "EnVap": -0.67, "ηa": 0.0, "ηb": 0.0
        }
    )
    chlorine: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Chlorine group contribution.",
        alias="-Cl",
        json_schema_extra={
            "category": "hologen increments", "group": "-Cl", "type": "normal", "id": "H2", "Tc": 0.0105, "Pc": -0.0049, "Vc": 58.0, "Tb": 38.13, "Tf": 13.55, "EnFo_IG": -71.55, "GiEnFo_IG": -64.31, "a": 33.3, "b": -0.0963, "c": 0.000187, "d": -9.96e-08, "EnFus": 2.515, "EnVap": 4.532, "ηa": 625.45, "ηb": -1.814
        }
    )
    bromine: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Bromine group contribution.",
        alias="-Br",
        json_schema_extra={
            "category": "hologen increments", "group": "-Br", "type": "normal", "id": "H3", "Tc": 0.0133, "Pc": 0.0057, "Vc": 71.0, "Tb": 66.86, "Tf": 43.43, "EnFo_IG": -29.48, "GiEnFo_IG": -38.06, "a": 28.6, "b": -0.0649, "c": 0.000136, "d": -7.45e-08, "EnFus": 3.603, "EnVap": 6.582, "ηa": 738.91, "ηb": -2.038
        }
    )
    iodine: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Iodine group contribution.",
        alias="-I",
        json_schema_extra={
            "category": "hologen increments", "group": "-I", "type": "normal", "id": "H4", "Tc": 0.0068, "Pc": -0.0034, "Vc": 97.0, "Tb": 93.84, "Tf": 41.69, "EnFo_IG": 21.06, "GiEnFo_IG": 5.74, "a": 32.1, "b": -0.0641, "c": 0.000126, "d": -6.87e-08, "EnFus": 2.724, "EnVap": 9.52, "ηa": 809.55, "ηb": -2.224
        }
    )
    alcohol_OH: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Alcohol OH group contribution.",
        alias="-OH @alcohol",
        json_schema_extra={
            "category": "oxygen increments", "group": "-OH @alcohol", "type": "alcohol", "id": "O1", "Tc": 0.0741, "Pc": 0.0112, "Vc": 28.0, "Tb": 92.88, "Tf": 44.45, "EnFo_IG": -208.04, "GiEnFo_IG": -189.2, "a": 25.7, "b": -0.0691, "c": 0.000177, "d": -9.88e-08, "EnFus": 2.406, "EnVap": 16.826, "ηa": 2173.72, "ηb": -5.057
        }
    )
    phenol_OH: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Phenol OH group contribution.",
        alias="-OH @phenol",
        json_schema_extra={
            "category": "oxygen increments", "group": "-OH @phenol", "type": "phenol", "id": "O2", "Tc": 0.024, "Pc": 0.0184, "Vc": -25.0, "Tb": 76.34, "Tf": 82.83, "EnFo_IG": -221.65, "GiEnFo_IG": -197.37, "a": -2.81, "b": 0.111, "c": -0.000116, "d": 4.94e-08, "EnFus": 4.49, "EnVap": 12.499, "ηa": 3018.17, "ηb": -7.314
        }
    )
    ether_non_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Ether O (non-ring) group contribution.",
        alias="-O- @non-ring",
        json_schema_extra={
            "category": "oxygen increments", "group": "-O- @non-ring", "type": "non-ring", "id": "O3", "Tc": 0.0168, "Pc": 0.0015, "Vc": 18.0, "Tb": 22.42, "Tf": 22.23, "EnFo_IG": -132.22, "GiEnFo_IG": -105.0, "a": 25.5, "b": -0.0632, "c": 0.000111, "d": -5.48e-08, "EnFus": 1.188, "EnVap": 2.41, "ηa": 122.09, "ηb": -0.386
        }
    )
    ether_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Ether O (ring) group contribution.",
        alias="-O- @ring",
        json_schema_extra={
            "category": "oxygen increments", "group": "-O- @ring", "type": "ring", "id": "O4", "Tc": 0.0098, "Pc": 0.0048, "Vc": 13.0, "Tb": 31.22, "Tf": 23.05, "EnFo_IG": -138.16, "GiEnFo_IG": -98.22, "a": 12.2, "b": -0.0126, "c": 6.03e-05, "d": -3.86e-08, "EnFus": 5.879, "EnVap": 4.682, "ηa": 440.24, "ηb": -0.953
        }
    )
    carbonyl_non_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Carbonyl C=O (non-ring) group contribution.",
        alias=">C=O @non-ring",
        json_schema_extra={
            "category": "oxygen increments", "group": ">C=O @non-ring", "type": "non-ring", "id": "O5", "Tc": 0.038, "Pc": 0.0031, "Vc": 62.0, "Tb": 76.75, "Tf": 61.2, "EnFo_IG": -133.22, "GiEnFo_IG": -120.5, "a": 6.45, "b": 0.067, "c": -3.57e-05, "d": 2.86e-09, "EnFus": 4.189, "EnVap": 8.972, "ηa": 340.35, "ηb": -0.35
        }
    )
    carbonyl_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Carbonyl C=O (ring) group contribution.",
        alias=">C=O @ring",
        json_schema_extra={
            "category": "oxygen increments", "group": ">C=O @ring", "type": "ring", "id": "O6", "Tc": 0.0284, "Pc": 0.0028, "Vc": 55.0, "Tb": 94.97, "Tf": 75.97, "EnFo_IG": -164.5, "GiEnFo_IG": -126.27, "a": 30.4, "b": -0.0829, "c": 0.000236, "d": -1.31e-07, "EnFus": 0.0, "EnVap": 6.645, "ηa": 0.0, "ηb": 0.0
        }
    )
    aldehyde: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Aldehyde O=CH- group contribution.",
        alias="O=CH-",
        json_schema_extra={
            "category": "oxygen increments", "group": "O=CH- @aldehyde", "type": "aldehyde", "id": "O7", "Tc": 0.0379, "Pc": 0.003, "Vc": 82.0, "Tb": 72.24, "Tf": 36.9, "EnFo_IG": -162.03, "GiEnFo_IG": -143.48, "a": 30.9, "b": -0.0336, "c": 0.00016, "d": -9.88e-08, "EnFus": 3.197, "EnVap": 9.093, "ηa": 740.92, "ηb": -1.713
        }
    )
    carboxylic_acid: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Carboxylic acid COOH group contribution.",
        alias="-COOH",
        json_schema_extra={
            "category": "oxygen increments", "group": "-COOH @acid", "type": "acid", "id": "O8", "Tc": 0.0791, "Pc": 0.0077, "Vc": 89.0, "Tb": 169.09, "Tf": 155.5, "EnFo_IG": -426.72, "GiEnFo_IG": -387.87, "a": 24.1, "b": 0.0427, "c": 8.04e-05, "d": -6.87e-08, "EnFus": 11.051, "EnVap": 19.537, "ηa": 1317.23, "ηb": -2.578
        }
    )
    ester: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Ester COO- group contribution.",
        alias="-COO-",
        json_schema_extra={
            "category": "oxygen increments", "group": "-COO- @ester", "type": "ester", "id": "O9", "Tc": 0.0481, "Pc": 0.0005, "Vc": 82.0, "Tb": 81.1, "Tf": 53.6, "EnFo_IG": -337.92, "GiEnFo_IG": -301.95, "a": 24.5, "b": 0.0402, "c": 4.02e-05, "d": -4.52e-08, "EnFus": 6.959, "EnVap": 9.633, "ηa": 483.88, "ηb": -0.966
        }
    )
    carbonyl_other: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Carbonyl =O (other) group contribution.",
        alias="=O @expect as above",
        json_schema_extra={
            "category": "oxygen increments", "group": "=O", "type": "expect as above", "id": "O10", "Tc": 0.0143, "Pc": 0.0101, "Vc": 36.0, "Tb": -10.5, "Tf": 2.08, "EnFo_IG": -247.61, "GiEnFo_IG": -250.83, "a": 6.82, "b": 0.0196, "c": 1.27e-05, "d": -1.78e-08, "EnFus": 3.624, "EnVap": 5.909, "ηa": 675.24, "ηb": -1.34
        }
    )
    primary_amine: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Primary amine NH2 group contribution.",
        alias="-NH2",
        json_schema_extra={
            "category": "nitrogen increments", "group": "-NH2", "type": "normal", "id": "N1", "Tc": 0.0243, "Pc": 0.0109, "Vc": 38.0, "Tb": 73.23, "Tf": 66.89, "EnFo_IG": -22.02, "GiEnFo_IG": 14.07, "a": 26.9, "b": -0.0412, "c": 0.000164, "d": -9.76e-08, "EnFus": 3.515, "EnVap": 10.788, "ηa": 0.0, "ηb": 0.0
        }
    )
    secondary_amine_non_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Secondary amine >NH (non-ring) group contribution.",
        alias=">NH @non-ring",
        json_schema_extra={
            "category": "nitrogen increments", "group": ">NH @non-ring", "type": "non-ring", "id": "N2", "Tc": 0.0295, "Pc": 0.0077, "Vc": 35.0, "Tb": 50.17, "Tf": 52.66, "EnFo_IG": 53.47, "GiEnFo_IG": 89.39, "a": -1.21, "b": 0.0762, "c": -4.86e-05, "d": 1.05e-08, "EnFus": 5.099, "EnVap": 6.436, "ηa": 0.0, "ηb": 0.0
        }
    )
    secondary_amine_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Secondary amine >NH (ring) group contribution.",
        alias=">NH @ring",
        json_schema_extra={
            "category": "nitrogen increments", "group": ">NH @ring", "type": "ring", "id": "N3", "Tc": 0.013, "Pc": 0.0114, "Vc": 29.0, "Tb": 52.82, "Tf": 101.51, "EnFo_IG": 31.65, "GiEnFo_IG": 75.61, "a": 11.8, "b": -0.023, "c": 0.000107, "d": -6.28e-08, "EnFus": 7.49, "EnVap": 6.93, "ηa": 0.0, "ηb": 0.0
        }
    )
    tertiary_amine_non_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Tertiary amine >N- (non-ring) group contribution.",
        alias=">N- @non-ring",
        json_schema_extra={
            "category": "nitrogen increments", "group": ">N- @non-ring", "type": "non-ring", "id": "N4", "Tc": 0.0169, "Pc": 0.0074, "Vc": 9.0, "Tb": 11.74, "Tf": 48.84, "EnFo_IG": 123.34, "GiEnFo_IG": 163.16, "a": -31.1, "b": 0.227, "c": -0.00032, "d": 1.46e-07, "EnFus": 4.703, "EnVap": 1.896, "ηa": 0.0, "ηb": 0.0
        }
    )
    imine_non_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Imine -N= (non-ring) group contribution.",
        alias="-N= @non-ring",
        json_schema_extra={
            "category": "nitrogen increments", "group": "-N= @non-ring", "type": "non-ring", "id": "N5", "Tc": 0.0255, "Pc": -0.0099, "Vc": 0.0, "Tb": 74.6, "Tf": 0.0, "EnFo_IG": 23.61, "GiEnFo_IG": 0.0, "a": 0.0, "b": 0.0, "c": 0.0, "d": 0.0, "EnFus": 0.0, "EnVap": 3.335, "ηa": 0.0, "ηb": 0.0
        }
    )
    imine_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Imine -N= (ring) group contribution.",
        alias="-N= @ring",
        json_schema_extra={
            "category": "nitrogen increments", "group": "-N= @ring", "type": "ring", "id": "N6", "Tc": 0.0085, "Pc": 0.0076, "Vc": 34.0, "Tb": 57.55, "Tf": 68.4, "EnFo_IG": 55.52, "GiEnFo_IG": 79.93, "a": 8.83, "b": -0.00384, "c": 4.35e-05, "d": -2.6e-08, "EnFus": 3.649, "EnVap": 6.528, "ηa": 0.0, "ηb": 0.0
        }
    )
    imine_secondary: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Secondary imine =NH group contribution.",
        alias="=NH",
        json_schema_extra={
            "category": "nitrogen increments", "group": "=NH", "type": "normal", "id": "N7", "Tc": 0.0, "Pc": 0.0, "Vc": 0.0, "Tb": 83.08, "Tf": 68.91, "EnFo_IG": 93.7, "GiEnFo_IG": 119.66, "a": 5.69, "b": -0.00412, "c": 0.000128, "d": -8.88e-08, "EnFus": 0.0, "EnVap": 12.169, "ηa": 0.0, "ηb": 0.0
        }
    )
    nitrile: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Nitrile CN group contribution.",
        alias="-CN",
        json_schema_extra={
            "category": "nitrogen increments", "group": "-CN", "type": "normal", "id": "N8", "Tc": 0.0496, "Pc": -0.0101, "Vc": 91.0, "Tb": 125.66, "Tf": 59.89, "EnFo_IG": 88.43, "GiEnFo_IG": 89.22, "a": 36.5, "b": -0.0733, "c": 0.000184, "d": -1.03e-07, "EnFus": 2.414, "EnVap": 12.851, "ηa": 0.0, "ηb": 0.0
        }
    )
    nitro: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Nitro NO2 group contribution.",
        alias="-NO2",
        json_schema_extra={
            "category": "nitrogen increments", "group": "-NO2", "type": "normal", "id": "N9", "Tc": 0.0437, "Pc": 0.0064, "Vc": 91.0, "Tb": 152.54, "Tf": 127.24, "EnFo_IG": -66.57, "GiEnFo_IG": -16.83, "a": 25.9, "b": -0.00374, "c": 0.000129, "d": -8.88e-08, "EnFus": 9.679, "EnVap": 16.738, "ηa": 0.0, "ηb": 0.0
        }
    )
    thiol: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Thiol SH group contribution.",
        alias="-SH",
        json_schema_extra={
            "category": "sulfur increments", "group": "-SH", "type": "normal", "id": "S1", "Tc": 0.0031, "Pc": 0.0084, "Vc": 63.0, "Tb": 63.56, "Tf": 20.09, "EnFo_IG": -17.33, "GiEnFo_IG": -22.99, "a": 35.3, "b": -0.0758, "c": 0.000185, "d": -1.03e-07, "EnFus": 2.36, "EnVap": 6.884, "ηa": 0.0, "ηb": 0.0
        }
    )
    thioether_non_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Thioether -S- (non-ring) group contribution.",
        alias="-S- @non-ring",
        json_schema_extra={
            "category": "sulfur increments", "group": "-S- @non-ring", "type": "non-ring", "id": "S2", "Tc": 0.0119, "Pc": 0.0049, "Vc": 54.0, "Tb": 68.78, "Tf": 34.4, "EnFo_IG": 41.87, "GiEnFo_IG": 33.12, "a": 19.6, "b": -0.00561, "c": 4.02e-05, "d": -2.76e-08, "EnFus": 4.13, "EnVap": 6.817, "ηa": 0.0, "ηb": 0.0
        }
    )
    thioether_ring: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        description="Thioether -S- (ring) group contribution.",
        alias="-S- @ring",
        json_schema_extra={
            "category": "sulfur increments", "group": "-S- @ring", "type": "ring", "id": "S3", "Tc": 0.0019, "Pc": 0.0051, "Vc": 38.0, "Tb": 52.1, "Tf": 79.93, "EnFo_IG": 39.1, "GiEnFo_IG": 27.76, "a": 16.7, "b": 0.00481, "c": 2.77e-05, "d": -2.11e-08, "EnFus": 1.557, "EnVap": 5.984, "ηa": 0.0, "ηb": 0.0
        }
    )


class JobackGroupData(BaseModel):
    """A class to represent a Joback group with its name and contribution data."""
    id: str
    name: str
    count: float
    data: Dict[str, str]


class JobackHeatCapacity(BaseModel):
    a: float = Field(..., description="Cp correlation parameter a")
    b: float = Field(..., description="Cp correlation parameter b")
    c: float = Field(..., description="Cp correlation parameter c")
    d: float = Field(..., description="Cp correlation parameter d")

    # optional: make parameters immutable after creation
    model_config = ConfigDict(frozen=True)

    def __call__(self, T: float) -> float:
        """
        Evaluate Cp at temperature T (K).
        """
        return (
            self.a - 37.93 +
            (self.b + 0.210) * T +
            (self.c - 3.91e-4) * T**2 +
            (self.d + 2.06e-7) * T**3
        )

    def Cp(self, T: float) -> float:
        """
        Alias method if you prefer Cp(T) instead of obj(T).
        """
        return self(T)


class JobackProp(TypedDict):
    value: float | None
    unit: str
    symbol: str


class JobackCalcProp(TypedDict):
    value: Callable[[float], float] | None
    unit: str
    symbol: str



# import libs
import logging
from typing import List, Tuple, Dict
# locals
from ..models import JobackGroupContributions

# NOTE: logger
logger = logging.getLogger(__name__)


def joback_group_contribution_info() -> Tuple[List[str], List[str]]:
    """
    Get the list of Joback group contribution.

    Returns
    -------
    Tuple[List[str],List[str]]
        A tuple containing two lists: field names and their corresponding aliases.
    """
    try:
        # NOTE: ids
        group_ids = []
        group_names = []

        # NOTE: alias
        for filed_name, field_info in JobackGroupContributions.model_fields.items():
            # get alias
            alias = field_info.alias

            # >> store alias
            if alias is not None:
                group_ids.append(alias)

            # >> also store field names
            group_names.append(filed_name)

        return group_names, group_ids
    except Exception as e:
        logger.error(f"Error retrieving Joback group contribution IDs: {e}")
        return [], []


def joback_group_contribution_names() -> List[str]:
    """
    Get the list of Joback group contribution names.

    Returns
    -------
    List[str]
        A list of Joback group contribution IDs.
    """
    try:
        # NOTE: keys
        group_ids = list(JobackGroupContributions.model_fields.keys())
        return group_ids
    except Exception as e:
        logger.error(f"Error retrieving Joback group contribution IDs: {e}")
        return []


def joback_group_contribution_ids() -> List[str]:
    """
    Get the list of Joback group contribution IDs.

    Returns
    -------
    List[str]
        A list of Joback group contribution IDs.
    """
    try:
        # NOTE: ids
        _, group_ids = joback_group_contribution_info()

        return group_ids
    except Exception as e:
        logger.error(f"Error retrieving Joback group contribution IDs: {e}")
        return []


def joback_group_contribution_category() -> Dict[str, List[Dict[str, str]]]:
    """
    Get the Joback group contribution categorized by their categories.

    Returns
    -------
    Dict[str, List[Dict[str, str]]]
        A dictionary categorizing group contributions by their categories.
    """
    try:
        # NOTE: category
        category = {}

        # NOTE: alias
        for filed_name, field_info in JobackGroupContributions.model_fields.items():
            # ! alias
            alias = field_info.alias
            # ! category
            cat = field_info.json_schema_extra or None

            # check
            if cat and isinstance(cat, dict):
                cat_value = cat.get("category", "Unknown")
            else:
                cat_value = "Unknown"

            # >> create category
            if cat_value not in category:
                category[cat_value] = []

            # >> store alias
            if alias is None:
                alias = "N/A"

            # store in category
            res_ = {
                "group": filed_name,
                "alias": alias
            }
            # >> append
            category[cat_value].append(res_)

        return category
    except Exception as e:
        logger.error(f"Error retrieving Joback group contribution IDs: {e}")
        return {}



# import libs
from pydantic import BaseModel, Field, field_validator
from typing import Optional, List, Any
import numpy as np


class AntoineFitResult(BaseModel):
    A: Optional[float] = None
    B: Optional[float] = None
    C: Optional[float] = None
    base: Optional[str] = None
    p_unit: str = "pa"
    T_unit_internal: str = "K"
    fit_in_log_space: bool = False
    success: bool = False
    message: str = ""
    cost: Optional[float] = None
    rmse_logP: Optional[float] = None
    mae_logP: Optional[float] = None
    r2_logP: Optional[float] = None
    rmse_P: Optional[float] = None
    mae_P: Optional[float] = None
    cov: Optional[Any] = None
    warnings: List[str] = Field(default_factory=list)
    Tmin_K: Optional[float] = None
    Tmax_K: Optional[float] = None
    loss: Optional[Any] = None
    f_scale: Optional[float] = None

    @field_validator("*", mode="before")
    def _convert_numpy_types(cls, v):
        if isinstance(v, (np.floating, np.integer)):
            return v.item()
        if isinstance(v, np.ndarray):
            return v.tolist()
        return v

    class Config:
        arbitrary_types_allowed = True
        orm_mode = True



# ref
from .ref import (
    GroupUnit,
    EstimatedProp,
)
# joback group contributions
from .jb import (
    JobackGroupContributions,
    JobackGroupData,
    JobackHeatCapacity,
    JobackProp,
    JobackCalcProp
)
# zabransky ruzicka
from .zr import (
    ZabranskyRuzickaGroupContributions,
    ZabranskyRuzickaGroupContributionsCorrections,
    ZabranskyRuzickaGroupData
)

__all__ = [
    "JobackGroupContributions",
    "GroupUnit",
    "JobackProp",
    "JobackCalcProp",
    "JobackGroupData",
    "JobackHeatCapacity",
    "ZabranskyRuzickaGroupContributions",
    "ZabranskyRuzickaGroupContributionsCorrections",
    "ZabranskyRuzickaGroupData",
    "EstimatedProp"
]



# import libs
import os
# local
from ..util import ReferenceLoader


class RefCore:
    '''
    Core class.
    '''

    def __init__(self):
        # init
        self.ReferenceLoader_ = ReferenceLoader()

        # load reference
        # self.ref_dict = self.load_ref()

    # def load_ref(self):
    #     '''
    #     Load reference files.
    #     '''
    #     # current file path
    #     current_path = os.path.dirname(os.path.abspath(__file__))
    #     # data folder path
    #     data_folder = os.path.join(
    #         current_path,
    #         '..',
    #         'data',
    #     )

    #     # load reference
    #     ref_dict = self.ReferenceLoader_.load_yml_ref(
    #         reference_name='sheet1.yml',
    #         reference_folder=data_folder,
    #     )
    #     # return reference
    #     return ref_dict



# import libs
import logging
from typing import Dict
from math import pow
import os
# locals
from ..models import (
    JobackGroupContributions,
    GroupUnit,
    JobackGroupData,
    JobackHeatCapacity,
    EstimatedProp
)
from ..util import ReferenceLoader
from ..configs import JOBACK_DATA_FILE, JOBACK_TABLE_COLUMN_GROUP

# NOTE: logger
logger = logging.getLogger(__name__)


class Joback:
    '''
    Joback method implementation.

    - The Joback group contribution method (often called Joback-Reid method) estimates 11 thermodynamic properties of organic compounds, such as boiling point, critical temperature, and heat of formation, by breaking molecules into 41 predefined functional groups and summing their contributions.
    - It assumes no interactions between groups, making it simple but less accurate for complex or aromatic compounds; estimates are generally within 5-15% error for many properties based on tested datasets.
    - To use it, identify and count the functional groups in the molecule (e.g., for ethanol: one -CH₃, one -CH₂-, one -OH @alcohol), then apply the summation formulas for the desired property—research suggests it's most reliable for mid-sized aliphatic organics, with potential inaccuracies for rings or halogens.
    - All 41 groups are categorized into non-ring, ring, halogen, oxygen, nitrogen, and sulfur types, with specific contribution values for each property.
    '''

    def __init__(
        self,
        group_contributions: JobackGroupContributions | Dict[str, float] | Dict[str, int],
        total_atoms_number: int,
    ):
        '''
        Initializes Joback method with group contributions.

        Parameters
        ----------
        group_contributions : JobackGroupContributions | Dict[str, float] | Dict[str, int]
            Group contributions for Joback method.
        '''
        # NOTE: group contributions
        self.group_contributions = group_contributions
        # NOTE: total atoms number
        self.total_atoms_number = total_atoms_number

        # SECTION: load Joback parameters
        self.joback_params = self.load_joback_parameters()

        # SECTION: get group contribution
        self.group_id = self._get_group_contribution()

        # SECTION: valid groups
        self.valid_groups = self._check_group_contributions()

    def __repr__(self) -> str:
        return f"""Joback Method with {len(self.group_id)} groups  \n
        Group Contributions: {self.group_contributions} """

    @property
    def group_contribution_idx(self):
        '''
        Returns the group contribution index.

        Returns
        -------
        group_id : dict
            Dictionary of group contributions.
        '''
        return self.group_id

    def load_joback_parameters(
        self,
    ):
        '''
        Loads Joback parameters from reference file.

        Returns
        -------
        joback_params : pd.DataFrame
            DataFrame of Joback parameters.
        '''
        try:
            # load reference file
            ref_loader = ReferenceLoader()

            # current folder
            current_folder = os.path.dirname(os.path.abspath(__file__))
            #  data folder
            data_folder = os.path.join(
                current_folder,
                '..',
                'data',
            )

            # NOTE: load CSV reference
            joback_df = ref_loader.load_csv_ref(
                reference_name=JOBACK_DATA_FILE,
                reference_folder=data_folder,
            )

            # return Joback parameters
            return joback_df
        except Exception as e:
            raise Exception("Loading Joback parameters failed!, ", e)

    def _get_group_contribution(
        self,
    ):
        '''
        Gets group contribution for a specific group.

        Parameters
        ----------
        None

        Returns
        -------
        contribution : dict
            Dictionary of group contributions for various properties.
        '''
        try:
            # filter Joback parameters for the specific group
            group_data = self.joback_params[
                JOBACK_TABLE_COLUMN_GROUP
            ]

            # extract contributions
            contribution = group_data.to_dict()

            return contribution
        except Exception as e:
            raise Exception(f"Getting contribution for group failed!, ", e)

    def list_available_groups(
        self,
    ):
        '''
        Lists all available Joback groups.

        Returns
        -------
        groups : list
            List of available Joback groups.
        '''
        try:
            # get list of available groups
            groups = self.joback_params[JOBACK_TABLE_COLUMN_GROUP].tolist()

            return groups
        except Exception as e:
            raise Exception("Listing available Joback groups failed!, ", e)

    def _get_group_contribution_data(
        self,
        group_name: str,
    ):
        '''
        Gets contribution data for a specific group.

        Parameters
        ----------
        group_name : str
            Name of the group.

        Returns
        -------
        group_data : dict
            Dictionary of contribution data for the group with header names as keys.
        '''
        try:
            # filter Joback parameters for the specific group
            group_data = self.joback_params[
                self.joback_params[JOBACK_TABLE_COLUMN_GROUP] == group_name.strip(
                )
            ]

            # convert to Series and then to dictionary with column headers as keys
            return {k: str(v) for k, v in group_data.iloc[0].to_dict().items()} if len(group_data) > 0 else {}
        except Exception as e:
            raise Exception(
                f"Getting contribution data for group {group_name} failed!, ", e)

    def _check_group_contributions(
            self
    ) -> Dict[str, JobackGroupData]:
        """
        Checks the validity of group contributions.
        """
        try:
            # NOTE: count occurrences
            valid_groups = {}

            # count each group
            if isinstance(self.group_contributions, dict):
                # ! dictionary type
                # iterate over group contributions
                for group_name, group_value in self.group_contributions.items():
                    # check if group exists in Joback parameters
                    if group_name in self.group_id.values():
                        # add to count dictionary
                        res_ = JobackGroupData(
                            id=group_name,
                            name=group_name,
                            count=float(group_value),
                            data=self._get_group_contribution_data(group_name)
                        )
                        # append
                        valid_groups[group_name] = res_
            elif isinstance(self.group_contributions, JobackGroupContributions):
                # ! dataclass type
                # iterate over dataclass fields
                for field_name, field_info in JobackGroupContributions.model_fields.items():
                    # get field alias (Joback group name)
                    alias = field_info.alias

                    # check if field is None
                    if alias is None:
                        continue

                    # value stored in the model instance = use field_name
                    group_unit: GroupUnit = getattr(
                        self.group_contributions,
                        field_name
                    )

                    # check if group value is greater than 0
                    if group_unit.value > 0:
                        # check if group exists in Joback parameters
                        if alias in self.group_id.values():
                            # add to count dictionary
                            res_ = JobackGroupData(
                                id=alias,
                                name=field_name,
                                count=float(group_unit.value),
                                data=self._get_group_contribution_data(alias)
                            )
                            # append
                            valid_groups[alias] = res_

            else:
                logger.error("Invalid type for group contributions!")
                return valid_groups

            return valid_groups
        except Exception as e:
            raise Exception("Checking group contributions failed!, ", e)

    def _calc_sigma(
            self
    ):
        '''
        Calculates sigma for all valid groups.

        '''
        try:
            # SECTION: calculate sigma
            # initialize critical temperature
            sigma = {
                'Tc': 0.0,
                'Pc': 0.0,
                'Vc': 0.0,
                'Tb': 0.0,
                'Tf': 0.0,
                'EnFo_IG': 0.0,
                'GiEnFo_IG': 0.0,
                'a': 0.0,
                'b': 0.0,
                'c': 0.0,
                'd': 0.0,
                'EnFus': 0.0,
                'EnVap': 0.0
            }

            # iterate over valid groups
            for group_name, group_info in self.valid_groups.items():
                # get group count
                group_count = group_info.count
                # get contribution data
                contribution_data = group_info.data
                # get contribution
                val_Tc = contribution_data.get('Tc', None)
                val_Pc = contribution_data.get('Pc', None)
                val_Vc = contribution_data.get('Vc', None)
                val_Tb = contribution_data.get('Tb', None)
                val_Tm = contribution_data.get('Tf', None)
                val_EnFo_IG = contribution_data.get('EnFo_IG', None)
                val_GiEnFo_IG = contribution_data.get('GiEnFo_IG', None)
                val_a = contribution_data.get('a', None)
                val_b = contribution_data.get('b', None)
                val_c = contribution_data.get('c', None)
                val_d = contribution_data.get('d', None)
                val_EnFus = contribution_data.get('EnFus', None)
                val_EnVap = contribution_data.get('EnVap', None)

                # check if contribution is valid
                if val_Tc is not None:
                    # update critical temperature
                    sigma['Tc'] += group_count * float(val_Tc)

                if val_Pc is not None:
                    # update critical temperature
                    sigma['Pc'] += group_count * float(val_Pc)

                if val_Vc is not None:
                    # update critical temperature
                    sigma['Vc'] += group_count * float(val_Vc)

                if val_Tb is not None:
                    # update critical temperature
                    sigma['Tb'] += group_count * float(val_Tb)

                if val_Tm is not None:
                    # update critical temperature
                    sigma['Tf'] += group_count * float(val_Tm)

                if val_EnFo_IG is not None:
                    # update critical temperature
                    sigma['EnFo_IG'] += group_count * float(val_EnFo_IG)

                if val_GiEnFo_IG is not None:
                    # update critical temperature
                    sigma['GiEnFo_IG'] += group_count * float(val_GiEnFo_IG)

                if val_a is not None:
                    # update critical temperature
                    sigma['a'] += group_count * float(val_a)

                if val_b is not None:
                    # update critical temperature
                    sigma['b'] += group_count * float(val_b)

                if val_c is not None:
                    # update critical temperature
                    sigma['c'] += group_count * float(val_c)

                if val_d is not None:
                    # update critical temperature
                    sigma['d'] += group_count * float(val_d)

                if val_EnFus is not None:
                    # update critical temperature
                    sigma['EnFus'] += group_count * float(val_EnFus)

                if val_EnVap is not None:
                    sigma['EnVap'] += group_count * float(val_EnVap)

            return sigma
        except Exception as e:
            raise Exception("Calculating critical temperature failed!, ", e)

    def _calc(
            self
    ):
        '''
        Calculates properties using Joback method.

        Returns
        -------
        properties : dict
            Dictionary of calculated properties.
        '''
        try:
            # SECTION: calculate properties
            properties: Dict[str, EstimatedProp] = {}

            # SECTION: calculate sigma
            sigma = self._calc_sigma()

            # NOTE: freezing point temperature
            properties['freezing_point_temperature'] = EstimatedProp(
                **self._calc_freezing_point_temperature(sigma)
            )

            # NOTE: boiling point temperature
            boiling_point_temp = self._calc_boiling_point_temperature(sigma)
            properties['boiling_point_temperature'] = EstimatedProp(
                **boiling_point_temp
            )

            # NOTE: critical temperature
            res_ = \
                self._calc_critical_temperature(
                    sigma,
                    boiling_point_temperature=boiling_point_temp['value'] if boiling_point_temp else None
                )
            properties['critical_temperature'] = EstimatedProp(
                **res_
            )

            # NOTE: critical pressure
            properties['critical_pressure'] = EstimatedProp(
                **self._calc_critical_pressure(sigma)
            )

            # NOTE: critical volume
            properties['critical_volume'] = EstimatedProp(
                **self._calc_critical_volume(sigma)
            )

            # NOTE: standard enthalpy of formation in ideal gas
            properties['standard_enthalpy_of_formation_ideal_gas'] = EstimatedProp(
                **self._calc_standard_enthalpy_of_formation_ideal_gas(sigma)
            )

            # NOTE: standard Gibbs energy of formation in ideal gas
            properties['standard_gibbs_energy_of_formation_ideal_gas'] = EstimatedProp(
                **self._calc_standard_gibbs_energy_of_formation_ideal_gas(sigma)
            )

            # NOTE: standard enthalpy of fusion
            properties['standard_enthalpy_of_fusion'] = EstimatedProp(
                **self._calc_standard_enthalpy_of_fusion(sigma)
            )

            # NOTE: standard enthalpy of vaporization
            properties['standard_enthalpy_of_vaporization'] = EstimatedProp(
                **self._calc_standard_enthalpy_of_vaporization(sigma)
            )

            # NOTE: heat capacity function
            properties['heat_capacity'] = EstimatedProp(
                **self._calc_heat_capacity(sigma)
            )

            return properties
        except Exception as e:
            raise Exception("Calculating properties failed!, ", e)

    def _calc_freezing_point_temperature(
            self,
            sigma: Dict[str, float]
    ):
        """
        Calculates the freezing point temperature using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        Tf : Dict
            Freezing point temperature (K).
        """
        try:
            # calc
            Tf = 122.5 + sigma['Tf']
            return {
                'value': Tf,
                'unit': 'K',
                'symbol': 'Tf'
            }
        except Exception as e:
            logger.error(
                f"Calculating freezing point temperature failed!, {e}")
            return {'value': None, 'unit': 'K', 'symbol': 'Tf'}

    def _calc_boiling_point_temperature(
            self,
            sigma: Dict[str, float]
    ):
        """
        Calculates the boiling point temperature using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        Tb : dict
            Boiling point temperature (K).
        """
        try:
            # calc
            Tb = 198.2 + sigma['Tb']
            return {
                'value': Tb,
                'unit': 'K',
                'symbol': 'Tb'
            }
        except Exception as e:
            logger.error(
                f"Calculating boiling point temperature failed!, {e}")
            return {'value': None, 'unit': 'K', 'symbol': 'Tb'}

    def _calc_critical_temperature(
            self,
            sigma: Dict[str, float],
            boiling_point_temperature: float | None
    ):
        """
        Calculates the critical temperature using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.
        boiling_point_temperature : float
            Boiling point temperature.

        Returns
        -------
        Tc : dict
            Critical temperature (K).
        """
        try:
            # check
            if boiling_point_temperature is None:
                logger.error(
                    f"Boiling point temperature is required for calculating critical temperature!")
                return {'value': None, 'unit': 'K', 'symbol': 'Tc'}

            # calc
            Tc = boiling_point_temperature / (
                0.584 +
                0.965 * sigma['Tc'] -
                sigma['Tc'] ** 2
            )
            return {
                'value': Tc,
                'unit': 'K',
                'symbol': 'Tc'
            }
        except Exception as e:
            logger.error(
                f"Calculating critical temperature failed!, {e}")
            return {'value': None, 'unit': 'K', 'symbol': 'Tc'}

    def _calc_critical_pressure(
            self,
            sigma: Dict[str, float],
    ):
        """
        Calculates the critical pressure using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        Pc: dict
            Critical pressure (bar).
        """
        try:
            # calc
            Pc = pow(
                0.113 +
                0.0032 * self.total_atoms_number -
                sigma['Pc'],
                -2
            )
            return {
                'value': Pc,
                'unit': 'bar',
                'symbol': 'Pc'
            }
        except Exception as e:
            logger.error(
                f"Calculating critical pressure failed!, {e}")
            return {'value': None, 'unit': 'bar', 'symbol': 'Pc'}

    def _calc_critical_volume(
            self,
            sigma: Dict[str, float],
    ):
        """
        Calculates the critical volume using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        Vc : dict
            Critical volume (cm³/mol).
        """
        try:
            # calc
            Vc = 17.5 + sigma['Vc']
            return {
                'value': Vc,
                'unit': 'cm3/mol',
                'symbol': 'Vc'
            }
        except Exception as e:
            logger.error(
                f"Calculating critical volume failed!, {e}")
            return {'value': None, 'unit': 'cm3/mol', 'symbol': 'Vc'}

    def _calc_standard_enthalpy_of_formation_ideal_gas(
            self,
            sigma: Dict[str, float],
    ):
        """
        Calculates the standard enthalpy of formation in ideal gas using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        EnFo_IG : dict
            Standard enthalpy of formation in ideal gas (kJ/mol).
        """
        try:
            # calc
            EnFo_IG = 68.29 + sigma['EnFo_IG']
            return {
                'value': EnFo_IG,
                'unit': 'kJ/mol',
                'symbol': 'EnFo_IG'
            }
        except Exception as e:
            logger.error(
                f"Calculating standard enthalpy of formation in ideal gas failed!, {e}")
            return {'value': None, 'unit': 'kJ/mol', 'symbol': 'EnFo_IG'}

    def _calc_standard_gibbs_energy_of_formation_ideal_gas(
            self,
            sigma: Dict[str, float],
    ):
        """
        Calculates the standard Gibbs energy of formation in ideal gas using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        GiEnFo_IG : dict
            Standard Gibbs energy of formation in ideal gas (kJ/mol).
        """
        try:
            # calc
            GiEnFo_IG = 53.88 + sigma['GiEnFo_IG']
            return {
                'value': GiEnFo_IG,
                'unit': 'kJ/mol',
                'symbol': 'GiEnFo_IG'
            }
        except Exception as e:
            logger.error(
                f"Calculating standard Gibbs energy of formation in ideal gas failed!, {e}")
            return {'value': None, 'unit': 'kJ/mol', 'symbol': 'GiEnFo_IG'}

    def _calc_standard_enthalpy_of_fusion(
            self,
            sigma: Dict[str, float],
    ):
        """
        Calculates the standard enthalpy of fusion using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        EnFus : dict
            Standard enthalpy of fusion (kJ/mol).
        """
        try:
            # calc
            EnFus = -0.88 + sigma['EnFus']
            return {
                'value': EnFus,
                'unit': 'kJ/mol',
                'symbol': 'EnFus'
            }
        except Exception as e:
            logger.error(
                f"Calculating standard enthalpy of fusion failed!, {e}")
            return {'value': None, 'unit': 'kJ/mol', 'symbol': 'EnFus'}

    def _calc_standard_enthalpy_of_vaporization(
            self,
            sigma: Dict[str, float],
    ):
        """
        Calculates the standard enthalpy of vaporization using Joback method.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        EnVap : dict
            Standard enthalpy of vaporization (kJ/mol).
        """
        try:
            # calc
            EnVap = 15.30 + sigma['EnVap']
            return {
                'value': EnVap,
                'unit': 'kJ/mol',
                'symbol': 'EnVap'
            }
        except Exception as e:
            logger.error(
                f"Calculating standard enthalpy of vaporization failed!, {e}")
            return {'value': None, 'unit': 'kJ/mol', 'symbol': 'EnVap'}

    def _calc_heat_capacity(
            self,
            sigma: Dict[str, float],
    ):
        """
        Create a placeholder for heat capacity calculation.

        Parameters
        ----------
        sigma : Dict[str, float]
            Dictionary of sigma values.

        Returns
        -------
        Cp_func : dict
            Heat capacity function Cp(T).
        """
        try:
            # create heat capacity function
            res_ = JobackHeatCapacity(
                a=sigma['a'],
                b=sigma['b'],
                c=sigma['c'],
                d=sigma['d']
            )

            return {
                'value': res_.Cp,
                'unit': 'J/mol·K',
                'symbol': 'Cp_IG'
            }
        except Exception as e:
            logger.error(
                f"Creating heat capacity function failed!, {e}")
            return {'value': None, 'unit': 'J/mol·K', 'symbol': 'Cp_IG'}



# import libs
import logging
from typing import List, Tuple
# locals
from ..models import (
    ZabranskyRuzickaGroupContributions,
    ZabranskyRuzickaGroupContributionsCorrections,
)

# NOTE: logger
logger = logging.getLogger(__name__)


def zabransky_ruzicka_group_contribution_info() -> Tuple[List[str], List[str]]:
    """
    Get the list of Zabransky-Ruzicka group contribution.

    Returns
    -------
    Tuple[List[str],List[str]]
        A tuple containing two lists: field names and their corresponding aliases.
    """
    try:
        # NOTE: ids
        group_ids = []
        group_names = []

        # NOTE: alias
        for filed_name, field_info in ZabranskyRuzickaGroupContributions.model_fields.items():
            # get alias
            alias = field_info.alias

            # >> store alias
            if alias is not None:
                group_ids.append(alias)

            # >> also store field names
            group_names.append(filed_name)

        return group_names, group_ids
    except Exception as e:
        logger.error(
            f"Error retrieving Zabransky-Ruzicka group contribution IDs: {e}")
        return [], []


def zabransky_ruzicka_group_contribution_names() -> List[str]:
    """
    Get the list of Zabransky-Ruzicka group contribution names.

    Returns
    -------
    List[str]
        A list of Zabransky-Ruzicka group contribution IDs.
    """
    try:
        # NOTE: keys
        group_ids = list(
            ZabranskyRuzickaGroupContributions.model_fields.keys())
        return group_ids
    except Exception as e:
        logger.error(
            f"Error retrieving Zabransky-Ruzicka group contribution IDs: {e}")
        return []


def zabransky_ruzicka_group_contribution_ids() -> List[str]:
    """
    Get the list of Zabransky-Ruzicka group contribution IDs.

    Returns
    -------
    List[str]
        A list of Zabransky-Ruzicka group contribution IDs.
    """
    try:
        # NOTE: alias
        group_ids = []
        for filed_name, field_info in ZabranskyRuzickaGroupContributions.model_fields.items():
            # get alias
            alias = field_info.alias

            # >> store alias
            if alias is not None:
                group_ids.append(alias)

        return group_ids
    except Exception as e:
        logger.error(
            f"Error retrieving Zabransky-Ruzicka group contribution IDs: {e}")
        return []


def zabransky_ruzicka_group_correction_ids() -> List[str]:
    """
    Get the list of Zabransky-Ruzicka group contribution correction IDs.

    Returns
    -------
    List[str]
        A list of Zabransky-Ruzicka group contribution correction IDs.
    """
    try:
        # NOTE: alias
        group_ids = []
        for filed_name, field_info in ZabranskyRuzickaGroupContributionsCorrections.model_fields.items():
            # get alias
            alias = field_info.alias

            # >> store alias
            if alias is not None:
                group_ids.append(alias)

        return group_ids
    except Exception as e:
        logger.error(
            f"Error retrieving Zabransky-Ruzicka group contribution correction IDs: {e}")
        return []


def zabransky_ruzicka_group_correction_names() -> List[str]:
    """
    Get the list of Zabransky-Ruzicka group contribution correction names.

    Returns
    -------
    List[str]
        A list of Zabransky-Ruzicka group contribution correction IDs.
    """
    try:
        # NOTE: keys
        group_ids = list(
            ZabranskyRuzickaGroupContributionsCorrections.model_fields.keys())
        return group_ids
    except Exception as e:
        logger.error(
            f"Error retrieving Zabransky-Ruzicka group contribution correction IDs: {e}")
        return []


def zabransky_ruzicka_group_correction_info() -> Tuple[List[str], List[str]]:
    """
    Get the list of Zabransky-Ruzicka group contribution corrections.

    Returns
    -------
    Tuple[List[str],List[str]]
        A tuple containing two lists: field names and their corresponding aliases.
    """
    try:
        # NOTE: ids
        group_ids = []
        group_names = []

        # NOTE: alias
        for filed_name, field_info in ZabranskyRuzickaGroupContributionsCorrections.model_fields.items():
            # get alias
            alias = field_info.alias

            # >> store alias
            if alias is not None:
                group_ids.append(alias)

            # >> also store field names
            group_names.append(filed_name)

        return group_names, group_ids
    except Exception as e:
        logger.error(
            f"Error retrieving Zabransky-Ruzicka group contribution correction IDs: {e}")
        return [], []



# import libs
from typing import Callable, TypedDict
from pydantic import BaseModel, Field

# NOTE: Quantity


class GroupUnit(BaseModel):
    """A class to represent a physical quantity with a value and unit."""
    value: float = Field(
        0,
        description="The numerical value of the quantity."
    )


class Eq(BaseModel):
    value: Callable[[float], float]
    unit: str
    symbol: str


class EstimatedProp(TypedDict):
    value: Callable[[float], float] | float | None
    unit: str
    symbol: str



# import libs
from typing import Optional, Dict
from pydantic import BaseModel, Field, ConfigDict
# local
from .ref import GroupUnit

# SECTION: Zabransky-Ruzicka Group Contributions


class ZabranskyRuzickaGroupContributions(BaseModel):
    """Zabransky-Ruzicka group contribution identifiers and their corresponding contributions to thermodynamic properties."""
    model_config = ConfigDict(populate_by_name=True, extra='forbid')

    C_H_3_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0),
        alias="C-(H)3(C)"
    )
    C_H_2_C_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)2")
    C_H_C_3: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)3")
    C_C_4: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)4")
    Cd_H_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(H)2")
    Cd_H_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(H)(C)")
    Cd_C_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(C)2")
    Cd_H_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(H)(Cd)")
    Cd_C_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(C)(Cd)")
    C_H_2_C_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(Cd)")
    C_H_2_Cd_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(Cd)2")
    C_H_C_2_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(Cd)")
    C_C_3_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(Cd)")
    Ct_H: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Ct-(H)")
    Ct_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Ct-(C)")
    Ct_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Ct-(CB)")
    Ca: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Ca")
    CB_H: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(H)")
    CB_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(C)")
    CB_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(Cd)")
    CB_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)")
    C_H_2_C_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(CB)")
    C_H_2_CB_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(CB)2")
    C_H_C_2_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(CB)")
    C_H_CB_3: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(CB)3")
    C_C_3_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(CB)")
    CBF_CBF_CB_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CBF-(CBF)(CB)2")
    CBF_CBF_2_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CBF-(CBF)2(CB)")
    CBF_CBF_3: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CBF-(CBF)3")
    C_H_3_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)3(O)")
    C_H_2_O_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(O)2")
    C_H_2_C_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(O)")
    C_H_2_C_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(CO)")
    C_H_2_CO_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(CO)2")
    C_H_2_CB_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(CB)(O)")
    C_H_C_2_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(O)")
    C_H_C_2_O_v2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(O)")
    C_H_C_2_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(CO)")
    C_C_2_O_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)2(O)2")
    C_C_3_O_ether: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(O)-ether")
    C_C_3_O_alcohol: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(O)-alcohol")
    C_C_3_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(CO)")
    Cd_H_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(H)(CO)")
    Cd_C_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(C)(CO)")
    CB_CB_2_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(O)")
    CB_CB_2_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(CO)")
    CO_H_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(H)(C)")
    CO_H_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(H)(Cd)")
    CO_C_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(C)2")
    CO_C_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(C)(Cd)")
    CO_C_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(C)(CB)")
    CO_H_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(H)(O)")
    CO_C_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(C)(O)")
    CO_Cd_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(Cd)(O)")
    CO_CB_O: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(CB)(O)")
    CO_O_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(O)(CO)")
    CO_O_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CO-(O)2")
    O_H_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(H)(C)")
    O_H_C_diol: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(H)(C)-diol")
    O_H_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(H)(CB)")
    O_H_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(H)(CO)")
    O_C_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(C)2")
    O_C_2_alcohol: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(C)2-alcohol")
    O_C_Cd: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(C)(Cd)")
    O_C_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(C)(CB)")
    O_CB_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(CB)2")
    O_C_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(C)(CO)")
    O_Cd_CO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(Cd)(CO)")
    C_H_3_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)3(N)")
    C_H_2_C_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(N)")
    C_H_C_2_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(N)")
    C_C_3_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(N)")
    CB_CB_2_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(N)")
    N_H_2_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(H)2(C)")
    N_H_2_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(H)2(N)")
    N_H_C_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(H)(C)2")
    N_H_C_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(H)(C)(CB)")
    N_H_C_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(H)(C)(N)")
    N_H_CB_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(H)(CB)(N)")
    N_H_CB_2_pyrrole: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(H)(CB)2-pyrrole")
    N_C_3: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(C)3")
    N_C_2_CB: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(C)2(CB)")
    N_C_2_N: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="N-(C)2(N)")
    NB_CB_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="NB-(CB)2")
    C_H_2_C_CN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(CN)")
    C_C_3_CN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(CN)")
    Cd_H_CN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(H)(CN)")
    CB_CB_2_CN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(CN)")
    CB_NCO_isocyanate: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(NCO)-isocyanate")
    CB_CB_2_NO2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(NO2)")
    C_H_2_C_NO2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(NO2)")
    O_C_NO2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="O-(C)(NO2)")
    C_H_2_C_Br: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(Br)")
    C_H_2_C_Cl: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(C1)")
    C_H_2_CB_Cl: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(CB)(C1)")
    C_H_2_C_F: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(F)")
    C_H_2_C_I: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(I)")
    C_H_C_2_Br: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(Br)")
    C_H_C_2_Cl: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(C1)")
    C_H_C_Cl_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)(C1)2")
    C_H_C_Cl_F: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)(C1)(F)")
    C_H_C_F_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)(F)2")
    C_C_Br_F_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)(Br)(F)2")
    C_C_3_Cl: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(C1)")
    C_C_2_Cl_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)2(C1)2")
    C_C_Cl_3: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)(C1)3")
    C_C_Cl_F_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)(C1)(F)2")
    C_C_Cl_2_F: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)(C1)2(F)")
    C_C_2_F_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)2(F)2")
    C_C_F_3: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)(F)3")
    C_CB_F_3: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(CB)(F)3")
    Cd_H_Cl: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(H)(C1)")
    Cd_Cl_F: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(C1)(F)")
    Cd_Cl_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(C1)2")
    Cd_F_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cd-(F)2")
    CB_CB_2_Br: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(Br)")
    CB_CB_2_Cl: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB(CB)2(C1)")
    CB_CB_2_F: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(F)")
    CB_CB_2_I: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(I)")
    C_H_3_S: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)3(S)")
    C_H_2_C_S: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)2(C)(S)")
    C_H_C_2_S: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(H)(C)2(S)")
    C_C_3_S: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="C-(C)3(S)")
    CB_CB_2_S: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="CB-(CB)2(S)")
    S_H_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="S-(H)(C)")
    S_C_2: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="S-(C)2")
    S_CB_C: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="S-(CB)(C)")
    S_CB_2_thiophene: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="S-(CB)2-thiophene")
    S_C_S: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="S-(C)(S)")


class ZabranskyRuzickaGroupContributionsCorrections(BaseModel):
    """Zabransky-Ruzicka group contribution correction identifiers and their corresponding corrections to thermodynamic properties."""
    model_config = ConfigDict(populate_by_name=True, extra='forbid')

    CYCLOPROPANEA: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cyclopropanea")
    CYCLOBUTANE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cyclobutane")
    CYCLOPENTANE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cyclopentane")
    CYCLOHEXANE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cyclohexane")
    CYCLOHEPTANEA: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cycloheptanea")
    CYCLOOCTANEA: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cyclooctanea")
    CYCLOPENTENE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cyclopentene")
    CYCLOHEXENE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cyclohexene")
    CYCLOHEXADIENE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cyclohexadiene")
    INDAN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Indan")
    G_1H_INDENEA: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="1H-indenea")
    HEXAHYDROINDAN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Hexahydroindan")
    TETRAHYDRONAPHTHALENEA: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Tetrahydronaphthalenea")
    DECAHYDRONAPHTHALENE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Decahydronaphthalene")
    ETHYLENEOXIDE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Ethyleneoxide")
    G_1_3_DIOXOLANE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="1,3-Dioxolane")
    FURAN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Furan")
    TETRAHYDROFURAN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Tetrahydrofuran")
    TETRAHYDROPYRAN: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Tetrahydropyran")
    PYRROLIDINE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Pyrrolidine")
    PIPERIDINE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Piperidine")
    THIACYCLOBUTANEA: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Thiacyclobutanea")
    THIACYCLOPENTANE: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Thiacyclopentane")
    THIACYCLOHEXANEA: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Thiacyclohexanea")
    CIS: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Cis-")
    TRANS: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Trans-")
    ORTHO: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Ortho-")
    META: Optional[GroupUnit] = Field(
        default_factory=lambda: GroupUnit(value=0), alias="Meta-")


class ZabranskyRuzickaGroupData(BaseModel):
    """A class to represent a Zabransky Ruzicka group with its name and contribution data."""
    id: str
    name: str
    count: float
    data: Dict[str, str]



# import libs

# SECTION: Joback method settings
JOBACK_DATA_FILE = 'joback.csv'

# NOTE: tables
JOBACK_TABLE_COLUMN_GROUP = 'Group'

# SECTION: zabransky-ruzicka method settings
ZABRANSKY_RUZICKA_DATA_FILE_1 = 'zabransky-ruzicka-1.csv'
ZABRANSKY_RUZICKA_DATA_FILE_2 = 'zabransky-ruzicka-2.csv'

# NOTE: tables
ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP = 'Group'

# NOTE: reference
ZABRANSKY_RUZICKA_REFERENCE = """Zábranský, Milan, and Vlastimil Růžička Jr. "Estimation of the heat capacities of organic liquids as a function of temperature using group additivity: an amendment." Journal of physical and chemical reference data 33.4 (2004): 1071-1081.
"""



from .joback import (
    joback_group_contribution_info,
    joback_group_contribution_ids,
    joback_group_contribution_names,
    joback_group_contribution_category
)

from .zabransky_ruzicka import (
    zabransky_ruzicka_group_contribution_info,
    zabransky_ruzicka_group_contribution_ids,
    zabransky_ruzicka_group_contribution_names,
    zabransky_ruzicka_group_correction_info,
    zabransky_ruzicka_group_correction_ids,
    zabransky_ruzicka_group_correction_names
)


__all__ = [
    "joback_group_contribution_ids",
    "joback_group_contribution_names",
    "joback_group_contribution_info",
    "joback_group_contribution_category",
    "zabransky_ruzicka_group_contribution_ids",
    "zabransky_ruzicka_group_contribution_names",
    "zabransky_ruzicka_group_contribution_info",
    "zabransky_ruzicka_group_correction_ids",
    "zabransky_ruzicka_group_correction_names",
    "zabransky_ruzicka_group_correction_info"
]



from .configs import (
    __version__,
    __author__,
    __description__,
    __email__,
    __license__,
)

from .app import (
    joback_calc,
    joback_prop_calc,
    joback_heat_capacity_calc,
    zabransky_ruzicka_calc
)

__all__ = [
    # config
    "__version__",
    "__author__",
    "__description__",
    "__email__",
    "__license__",
    # app
    "joback_calc",
    "joback_prop_calc",
    "joback_heat_capacity_calc",
    "zabransky_ruzicka_calc",
]



# import libs
import logging
from typing import List, Tuple, Optional, Literal
import numpy as np
from pythermodb_settings.models import Temperature, Pressure
from pathlib import Path
# local
from ..core import Antoine
from ..util import normalize_unit
from ..models.antoine import AntoineFitResult

# NOTE: set up logger
logger = logging.getLogger(__name__)


def estimate_coefficients(
    temperatures: List[Temperature],
    pressures: List[Pressure],
    *,
    base: Literal['log10', 'ln'] = "log10",
    fit_in_log_space: bool = True,
    weights: Optional[np.ndarray] = None,
    x0: Optional[Tuple[float, float, float]] = None,
    bounds: Optional[
        Tuple[
            Tuple[float, float, float],
            Tuple[float, float, float]
        ]
    ] = None,
    max_nfev: int = 5000,
    validate: bool = True,
    min_margin_kelvin: float = 1.0,
    # robust options
    loss: Literal['linear', 'soft_l1', 'huber', 'cauchy', 'arctan'] = "linear",
    f_scale: Optional[float] = None,
) -> Optional[AntoineFitResult]:
    """
    Estimate Antoine coefficients from experimental data to fit the Antoine equation as follows:

        log10(P) = A - B / (T + C)   (if base is "log10")
        ln(P)   = A - B / (T + C)    (if base is "ln")

    Parameters
    ----------
    temperatures : List[Temperature]
        List of Temperature models containing temperature values and units.
    pressures : List[Pressure]
        List of Pressure models containing pressure values and units.
    base : str, optional
        Logarithm base used in the Antoine equation ('log10' or 'ln'), by default "log10".
    fit_in_log_space : bool, optional
        Whether to perform fitting in logarithmic space, by default True.
    weights : Optional[np.ndarray], optional
        Optional weights for each data point, by default None.
    x0 : Optional[Tuple[float, float, float]], optional
        Initial guess for coefficients (A, B, C), by default None.
    bounds : Optional[Tuple[Tuple[float, float, float], Tuple[float, float, float]]], optional
        Bounds for coefficients (A, B, C), by default None.
    max_nfev : int, optional
        Maximum number of function evaluations for the optimizer, by default 5000.
    validate : bool, optional
        Whether to perform validation checks on the fitted coefficients, by default True.
    min_margin_kelvin : float, optional
            Minimum margin for (T + C) to avoid denominator issues (default 1.0 K).
    loss : str, Literal['linear', 'soft_l1', 'huber', 'cauchy', 'arctan'], optional
        Loss function for robust fitting: 'linear', 'soft_l1', 'huber', 'cauchy', 'arctan' (default 'linear').
    f_scale : Optional[float], optional
        Scaling parameter for robust loss (default None = auto).

    Returns
    -------
    Optional[AntoineFitResult]
        An AntoineFitResult model containing fitted coefficients and metrics, or None if estimation fails defined as:
        - A: float
            Fitted coefficient A.
        - B: float
            Fitted coefficient B.
        - C: float
            Fitted coefficient C.
        - base: str
            Logarithm base used in the Antoine equation.
        - p_unit: str
            Unit of pressure used.
        - T_unit_internal: str
            Internal unit of temperature used.
        - fit_in_log_space: bool
            Whether fitting was done in logarithmic space.
        - success: bool
            Whether the fitting was successful.
        - message: str
            Message regarding the fitting process.
        - cost: float
            Final cost value from the fitting.
        - rmse_logP: float
            Root Mean Square Error in log space.
        - mae_logP: float
            Mean Absolute Error in log space.
        - r2_logP: float
            R-squared value in log space.
        - rmse_P: float
            Root Mean Square Error in original pressure space.
        - mae_P: float
            Mean Absolute Error in original pressure space.
        - cov: Any
            Covariance matrix of the fitted parameters.
        - warnings: List[str]
            List of warnings generated during fitting.
        - Tmin_K: float
            Minimum temperature in Kelvin from the data.
        - Tmax_K: float
            Maximum temperature in Kelvin from the data.
        - loss: Any
            Loss function used.
        - f_scale: float
            Scaling parameter used for robust fitting.
    """
    try:
        # SECTION: Validate inputs
        # NOTE: check lengths
        if len(temperatures) != len(pressures):
            logger.error("Lengths of temperatures and pressures do not match.")
            return None

        # >> check types
        if not all(isinstance(t, Temperature) for t in temperatures):
            logger.error(
                "All items in temperatures must be of type Temperature."
            )
            return None

        if not all(isinstance(p, Pressure) for p in pressures):
            logger.error("All items in pressures must be of type Pressure.")
            return None

        # SECTION: Normalize units
        # NOTE: normalize temperatures to Kelvin
        norm_temperature = normalize_unit(
            data=temperatures,
            to_unit="K",
            valid_from=["C", "F", "K", "R"]
        )

        norm_pressures = normalize_unit(
            data=pressures,
            to_unit="bar",
            valid_from=["Pa", "kPa", "bar", "atm", "psi"]
        )

        if not norm_temperature or not norm_pressures:
            logger.error("Normalization of units failed.")
            return None

        # SECTION: Extract values
        temp_values = norm_temperature.get('data', [])
        # >> to array
        temp_values = np.array(temp_values)

        pres_values = norm_pressures.get('data', [])
        # >> to array
        pres_values = np.array(pres_values)

        if len(temp_values) != len(pres_values):
            logger.error(
                "Normalized temperature and pressure data lengths do not match.")
            return None

        # SECTION: Estimate coefficients
        antoine_model = Antoine.fit_antoine(
            T_data=temp_values,
            P_data=pres_values,
            base=base,
            T_unit="K",
            p_unit="Pa",
            fit_in_log_space=fit_in_log_space,
            weights=weights,
            x0=x0,
            bounds=bounds,
            max_nfev=max_nfev,
            validate=validate,
            min_margin_kelvin=min_margin_kelvin,
            loss=loss,
            f_scale=f_scale,
        )

        # >> return result model
        return AntoineFitResult(**antoine_model)
    except Exception as e:
        logger.exception(
            f"An error occurred during coefficient estimation: {e}")
        return None


def estimate_coefficients_from_experimental_data(
    experimental_data: str | Path,
    *,
    temperature_unit: Literal['K', 'C', 'F', 'R'] = 'K',
    pressure_unit: Literal['Pa', 'kPa', 'bar', 'atm', 'psi'] = 'Pa',
    base: str = "log10",
    fit_in_log_space: bool = True,
    weights: Optional[np.ndarray] = None,
    x0: Optional[Tuple[float, float, float]] = None,
    bounds: Optional[
        Tuple[
            Tuple[float, float, float],
            Tuple[float, float, float]
        ]
    ] = None,
    max_nfev: int = 5000,
    validate: bool = True,
    min_margin_kelvin: float = 1.0,
    # robust options
    loss: str = "linear",
    f_scale: Optional[float] = None,
) -> Optional[AntoineFitResult]:
    """
    Estimate Antoine coefficients from experimental data to fit the Antoine equation as follows:

        log10(P) = A - B / (T + C)   (if base is "log10")
        ln(P)   = A - B / (T + C)    (if base is "ln")

    Parameters
    ----------
    experimental_data : str | Path
        Path to experimental data file (CSV, JSON, etc.) containing temperature and pressure data.
    base : str, optional
        Logarithm base used in the Antoine equation ('log10' or 'ln'), by default "log10".
    fit_in_log_space : bool, optional
        Whether to perform fitting in logarithmic space, by default True.
    weights : Optional[np.ndarray], optional
        Optional weights for each data point, by default None.
    x0 : Optional[Tuple[float, float, float]], optional
        Initial guess for coefficients (A, B, C), by default None.
    bounds : Optional[Tuple[Tuple[float, float, float], Tuple[float, float, float]]], optional
        Bounds for coefficients (A, B, C), by default None.
    max_nfev : int, optional
        Maximum number of function evaluations for the optimizer, by default 5000.
    validate : bool, optional
        Whether to perform validation checks on the fitted coefficients, by default True.
    min_margin_kelvin : float, optional
        Minimum margin in Kelvin to ensure valid temperature range, by default 1.0.
    loss : str, optional
        Loss function for robust fitting, by default "linear".
    f_scale : Optional[float], optional
        Scale parameter for robust fitting, by default None.

    Returns
    -------
    Optional[AntoineFitResult]
        An AntoineFitResult model containing fitted coefficients and metrics, or None if estimation fails defined as:
        - A: float
            Fitted coefficient A.
        - B: float
            Fitted coefficient B.
        - C: float
            Fitted coefficient C.
        - base: str
            Logarithm base used in the Antoine equation.
        - p_unit: str
            Unit of pressure used.
        - T_unit_internal: str
            Internal unit of temperature used.
        - fit_in_log_space: bool
            Whether fitting was done in logarithmic space.
        - success: bool
            Whether the fitting was successful.
        - message: str
            Message regarding the fitting process.
        - cost: float
            Final cost value from the fitting.
        - rmse_logP: float
            Root Mean Square Error in log space.
        - mae_logP: float
            Mean Absolute Error in log space.
        - r2_logP: float
            R-squared value in log space.
        - rmse_P: float
            Root Mean Square Error in original pressure space.
        - mae_P: float
            Mean Absolute Error in original pressure space.
        - cov: Any
            Covariance matrix of the fitted parameters.
        - warnings: List[str]
            List of warnings generated during fitting.
        - Tmin_K: float
            Minimum temperature in Kelvin from the data.
        - Tmax_K: float
            Maximum temperature in Kelvin from the data.
        - loss: Any
            Loss function used.
        - f_scale: float
            Scaling parameter used for robust fitting.
    """
    try:
        # SECTION: Check inputs
        # >> check file existence
        data_path = Path(experimental_data)

        if not data_path.exists() or not data_path.is_file():
            logger.error(
                f"Experimental data file does not exist: {experimental_data}")
            return None

        # SECTION: Load experimental data by Antoine class method
        # >> load data
        # ! temperature in K, pressure in bar
        T_data, P_data = Antoine.load_experimental_data(
            experimental_data=data_path,
            T_unit=temperature_unit,
            P_unit=pressure_unit,
        )

        if T_data is None or P_data is None:
            logger.error(
                "Failed to load experimental data from the provided file.")
            return None

        # SECTION: Estimate coefficients
        antoine_model = Antoine.fit_antoine(
            T_data=T_data,
            P_data=P_data,
            base=base,
            T_unit="K",
            p_unit="Pa",
            fit_in_log_space=fit_in_log_space,
            weights=weights,
            x0=x0,
            bounds=bounds,
            max_nfev=max_nfev,
            validate=validate,
            min_margin_kelvin=min_margin_kelvin,
            loss=loss,
            f_scale=f_scale,
        )

        # >> return result model
        return AntoineFitResult(**antoine_model)
    except Exception as e:
        logger.exception(
            f"An error occurred during coefficient estimation from experimental data: {e}")
        return None



# import libs
import logging
import os
import yaml
import pandas as pd

# NOTE: logger
logger = logging.getLogger(__name__)


class ReferenceLoader:
    '''
    Used to load reference files.
    '''

    def __init__(self):
        pass

    def load_yml_ref(
        self,
        reference_name,
        reference_folder,
    ):
        '''
        Loads YAML reference file.

        Parameters
        ----------
        reference_name : str
            Name of the reference file.
        reference_folder : str
            Folder of the reference file.

        Returns
        -------
        ref_dict : dict
            Dictionary of the reference file.
        '''
        try:
            # reference path
            reference_path = os.path.join(
                reference_folder,
                reference_name
            )
            # load reference file
            with open(reference_path, 'r') as f:
                ref_dict = yaml.safe_load(f)

            # return reference dictionary
            return ref_dict
        except Exception as e:
            raise Exception("Loading reference failed!, ", e)

    def load_csv_ref(
        self,
        reference_name: str,
        reference_folder: str,
    ) -> pd.DataFrame:
        '''
        Loads CSV reference file.

        Parameters
        ----------
        reference_name : str
            Name of the reference file.
        reference_folder : str
            Folder of the reference file.

        Returns
        -------
        ref_df : pd.DataFrame
            DataFrame of the reference file.
        '''
        try:
            # reference path
            reference_path = os.path.join(
                reference_folder,
                reference_name
            )
            # load reference file
            ref_df = pd.read_csv(reference_path)

            # return reference DataFrame
            return ref_df
        except Exception as e:
            raise Exception("Loading reference failed!, ", e)



# import libs
from typing import Optional, Dict
from pydantic import BaseModel, Field, ConfigDict
# local
from .ref import GroupUnit

# SECTION: Constantinou-Gani Group Contribution


class ConstantinouGaniGroupContribution(BaseModel):
    """Constantinou-Gani Group Contribution Model Parameters"""

    model_config = ConfigDict(
        title="Constantinou-Gani Group Contribution Model Parameters",
        populate_by_name=True,
        extra="forbid",
    )



# import libs
import logging
from typing import Dict, Optional
# locals
from .models import (
    JobackGroupContributions,
    ZabranskyRuzickaGroupContributions,
    ZabranskyRuzickaGroupContributionsCorrections,
    EstimatedProp,
    JobackProp,
    JobackCalcProp
)
from .core import Joback, ZabranskyRuzicka


# NOTE: logger
logger = logging.getLogger(__name__)


# SECTION: Joback Group Contributions

def joback_calc(
    groups: JobackGroupContributions | Dict[str, int] | Dict[str, float],
    total_atoms_number: int
) -> Optional[Dict[str, EstimatedProp]]:
    """
    Using Joback method to calculate thermodynamic properties including

    Parameters
    ----------
    groups : JobackGroupContributions | Dict[str, float] | Dict[str, int]
        Group contributions for Joback method.
    total_atoms_number : int
        Total number of atoms in the molecule.

    Returns
    -------
    Dict[str, EstimatedProp] | None
        A dictionary containing calculated thermodynamic properties.

    Notes
    -----
    You can use either the field names or their aliases to specify group contributions. For example:
    - methyl has alias '-CH3'
    - methylene has alias '-CH2- @non-ring'
    - imine_non_ring has alias '-N= @non-ring'
    - imine_ring has alias '-N= @ring'
    """
    try:
        # SECTION: initialize Joback method
        Joback_ = Joback(
            group_contributions=groups,
            total_atoms_number=total_atoms_number
        )

        # NOTE: calculate properties
        return Joback_._calc()
    except Exception as e:
        logger.error(f"Error in Joback calculation: {e}")
        return None


def joback_prop_calc(
    groups: JobackGroupContributions | Dict[str, int] | Dict[str, float],
    total_atoms_number: int,
) -> Optional[Dict[str, JobackProp]]:
    """
    Using Joback method to calculate a only thermodynamic property.

    Parameters
    ----------
    groups : JobackGroupContributions | Dict[str, float] | Dict[str, int]
        Group contributions for Joback method.
    total_atoms_number : int
        Total number of atoms in the molecule.

    Returns
    -------
    Dict[str, JobackProp] | None
        A dictionary containing the calculated thermodynamic property.

    Notes
    -----
    You can use either the field names or their aliases to specify group contributions.
    """
    try:
        # SECTION: initialize Joback method
        joback_calc_ = joback_calc(
            groups=groups,
            total_atoms_number=total_atoms_number
        )

        # SECTION: return property (not callable ones in value)
        # NOTE: initialize result
        prop_result: Dict[str, JobackProp] = {}

        # NOTE: iterate over calculated properties
        if joback_calc_:
            for prop_name, prop_data in joback_calc_.items():
                value = prop_data['value']
                if callable(value):
                    continue

                # >> add to result
                prop_result[prop_name] = JobackProp(
                    value=value,
                    unit=prop_data['unit'],
                    symbol=prop_data['symbol']
                )

        return prop_result
    except Exception as e:
        logger.error(f"Error in Joback property calculation: {e}")
        return None


def joback_heat_capacity_calc(
    groups: JobackGroupContributions | Dict[str, int] | Dict[str, float],
    total_atoms_number: int,
) -> Optional[JobackCalcProp]:
    """
    Using Joback method to retrieve heat capacity equation (function of temperature [K]).

    Parameters
    ----------
    groups : JobackGroupContributions | Dict[str, float] | Dict[str, int]
        Group contributions for Joback method.
    total_atoms_number : int
        Total number of atoms in the molecule.

    Returns
    -------
    JobackCalcProp | None
        A dictionary containing calculated heat capacity property as:
        - value: equation to calculate heat capacity.
        - units: units of the heat capacity (J/mol.K).
        - symbol: symbol representing heat capacity (Cp_LIQ).

    Notes
    -----
    You can use either the field names or their aliases to specify group contributions.
    """
    try:
        # SECTION: initialize Joback method
        joback_calc_ = joback_calc(
            groups=groups,
            total_atoms_number=total_atoms_number
        )

        # SECTION: return heat capacity property
        if joback_calc_ and 'heat_capacity' in joback_calc_:
            heat_capacity_data = joback_calc_['heat_capacity']

            # check type
            if not callable(heat_capacity_data['value']):
                return None

            # >> return heat capacity
            return JobackCalcProp(
                value=heat_capacity_data['value'],
                unit=heat_capacity_data['unit'],
                symbol=heat_capacity_data['symbol']
            )

        return None
    except Exception as e:
        logger.error(f"Error in Joback heat capacity calculation: {e}")
        return None


# SECTION: Zabransky-Ruzicka Group Contributions


def zabransky_ruzicka_calc(
    group_contributions: ZabranskyRuzickaGroupContributions | Dict[str, float] | Dict[str, int],
    group_corrections: Optional[
        ZabranskyRuzickaGroupContributionsCorrections |
        Dict[str, float] |
        Dict[str, int]
    ] = None
) -> Optional[EstimatedProp]:
    """
    Using Zabransky-Ruzicka method to calculate thermodynamic properties including

    Parameters
    ----------
    group_contributions : ZabranskyRuzickaGroupContributions | Dict[str, float] | Dict[str, int]
        Group contributions for Zabransky-Ruzicka method.
    group_corrections : Optional[ZabranskyRuzickaGroupContributionsCorrections | Dict[str, float] | Dict[str, int]]
        Group correction contributions for Zabransky-Ruzicka method.

    Returns
    -------
    EstimatedProp | None
        A dictionary containing calculated thermodynamic properties as:
        - value: equation to calculate heat capacity.
        - units: units of the heat capacity (J/mol.K).
        - symbol: symbol representing heat capacity (Cp_LIQ).

    Notes
    -----
    You can use either the field names or their aliases to specify group contributions. For example:
    - C-(H)3(C)
    - C-(H)2(C)2
    - C-(H)(C)3

    """
    try:
        # SECTION: initialize Zabransky-Ruzicka method
        ZabranskyRuzicka_ = ZabranskyRuzicka(
            group_contributions=group_contributions,
            group_corrections=group_corrections
        )

        # NOTE: calculate properties
        return ZabranskyRuzicka_._calc()
    except Exception as e:
        logger.error(f"Error in Zabransky-Ruzicka calculation: {e}")
        return None



# import libs
import logging
from typing import Dict, Any, Literal, Optional
from math import pow
import os
# locals
from ..models import (
    GroupUnit,
    ZabranskyRuzickaGroupContributions,
    ZabranskyRuzickaGroupContributionsCorrections,
    ZabranskyRuzickaGroupData,
    EstimatedProp
)
from ..util import ReferenceLoader
from ..configs import (
    ZABRANSKY_RUZICKA_DATA_FILE_1,
    ZABRANSKY_RUZICKA_DATA_FILE_2,
    ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP,
    ZABRANSKY_RUZICKA_REFERENCE
)

# NOTE: logger
logger = logging.getLogger(__name__)


class ZabranskyRuzicka:
    '''
    Zabransky-Ruzicka method to estimate thermodynamic properties.

    A second-order group additivity approach for estimating the temperature-dependent heat capacity of pure organic liquids in the range from melting point to normal boiling point.
    '''

    def __init__(
        self,
        group_contributions: ZabranskyRuzickaGroupContributions | Dict[str, float] | Dict[str, int],
        group_corrections:  Optional[
            ZabranskyRuzickaGroupContributionsCorrections |
            Dict[str, float] |
            Dict[str, int]
        ] = None
    ):
        '''
        Initializes the Zabransky-Ruzicka method.

        Parameters
        ----------
        group_contributions : ZabranskyRuzickaGroupContributions | Dict[str, float] | Dict[str, int]
            Group contributions for the compound.
        group_corrections : ZabranskyRuzickaGroupContributionsCorrections | Dict[
            str, float] | Dict[str, int] | None
            Group corrections for the compound.
        '''
        # NOTE: group contributions
        self.group_contributions = group_contributions
        self.group_corrections = group_corrections if group_corrections else {}

        # SECTION: load parameters from reference file
        self.params, self.corrections = self.load_parameters()

        # NOTE: get group contribution (from reference)
        self.group_id = self._get_group_contribution()
        # NOTE: get correction contribution (from reference)
        self.correction_id = self._get_correction_contribution()

        # SECTION: valid groups
        # NOTE: check group contributions
        self.valid_group_contribution = self._check_group_contributions(
            group_x=self.group_contributions,
            group_id=self.group_id
        )
        # NOTE: check group corrections
        self.valid_group_corrections = self._check_group_contributions(
            group_x=self.group_corrections,
            group_id=self.correction_id
        )

        # >> merge valid groups
        self.valid_groups = {
            **self.valid_group_contribution,
            **self.valid_group_corrections
        }

    def __repr__(self) -> str:
        return f"""Zabransky Ruzicka Method: \n
        Group Contributions: {self.group_contributions} \n
        Group Corrections: {self.group_corrections}  \n
        Reference: {ZABRANSKY_RUZICKA_REFERENCE}
        """

    @property
    def group_contribution_idx(self):
        '''
        Returns the group contribution index.

        Returns
        -------
        group_id : dict
            Dictionary of group contributions.
        '''
        return self.group_id

    def load_parameters(
        self,
    ):
        '''
        Loads parameters from reference file.

        Returns
        -------
        params_df, corrections_df : pd.DataFrame, pd.DataFrame
            DataFrames of parameters and corrections.
        '''
        try:
            # load reference file
            ref_loader = ReferenceLoader()

            # current folder
            current_folder = os.path.dirname(os.path.abspath(__file__))
            #  data folder
            data_folder = os.path.join(
                current_folder,
                '..',
                'data',
            )

            # NOTE: load CSV reference
            params_df = ref_loader.load_csv_ref(
                reference_name=ZABRANSKY_RUZICKA_DATA_FILE_1,
                reference_folder=data_folder,
            )

            # NOTE: corrections
            corrections_df = ref_loader.load_csv_ref(
                reference_name=ZABRANSKY_RUZICKA_DATA_FILE_2,
                reference_folder=data_folder,
            )

            # return parameters dataframe
            return params_df, corrections_df
        except Exception as e:
            raise Exception("Loading  parameters failed!, ", e)

    def _get_group_contribution(
        self,
    ):
        '''
        Gets group contribution for a specific group.

        Parameters
        ----------
        None

        Returns
        -------
        contribution : dict
            Dictionary of group contributions for various properties.
        '''
        try:
            # filter  parameters for the specific group
            group_data = self.params[
                ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP
            ]

            # extract contributions
            contribution = group_data.to_dict()

            return contribution
        except Exception as e:
            raise Exception(f"Getting contribution for group failed!, ", e)

    def _get_correction_contribution(
        self,
    ):
        '''
        Gets correction contribution for a specific group.

        Parameters
        ----------
        None

        Returns
        -------
        contribution : dict
            Dictionary of correction contributions for various properties.
        '''
        try:
            # filter  corrections for the specific group
            correction_data = self.corrections[
                ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP
            ]

            # extract contributions
            contribution = correction_data.to_dict()

            return contribution
        except Exception as e:
            raise Exception(
                f"Getting correction contribution for group failed!, ", e)

    def list_available_groups(
        self,
    ):
        '''
        Lists all available  groups.

        Returns
        -------
        groups : list
            List of available  groups.
        '''
        try:
            # get list of available groups
            groups = self.params[ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP].tolist()

            return groups
        except Exception as e:
            raise Exception("Listing available  groups failed!, ", e)

    def _get_group_x_data(
        self,
        group_name: str,
        group_mode: Literal['contribution', 'correction'] = 'contribution'
    ):
        '''
        Gets contribution data for a specific group.

        Parameters
        ----------
        group_name : str
            Name of the group.
        group_mode : Literal['contribution', 'correction'], optional
            Mode of the group, by default 'contribution'.

        Returns
        -------
        group_data : dict
            Dictionary of contribution data for the group with header names as keys.
        '''
        try:
            # filter  parameters for the specific group
            if group_mode == 'contribution':
                group_data = self.params[
                    self.params[ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP] == group_name.strip(
                    )
                ]
            elif group_mode == 'correction':
                group_data = self.corrections[
                    self.corrections[ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP] == group_name.strip(
                    )
                ]
            else:
                return {}

            # convert to Series and then to dictionary with column headers as keys
            return {k: str(v) for k, v in group_data.iloc[0].to_dict().items()} if len(group_data) > 0 else {}
        except Exception as e:
            raise Exception(
                f"Getting contribution data for group {group_name} failed!, ", e)

    def _check_group_contributions(
            self,
            group_x: Dict[str, Any] | ZabranskyRuzickaGroupContributions | ZabranskyRuzickaGroupContributionsCorrections,
            group_id: Dict[str, Any]
    ) -> Dict[str, ZabranskyRuzickaGroupData]:
        """
        Checks the validity of group contributions.

        Parameters
        ----------
        group_x : Dict[str, Any] | ZabranskyRuzickaGroupContributions | ZabranskyRuzickaGroupContributionsCorrections
            Dictionary or dataclass of group contributions.
        group_id : Dict[str, Any]
            Dictionary of group contributions.
        """
        try:
            # NOTE: count occurrences
            valid_groups = {}

            # count each group
            if isinstance(group_x, dict):
                # ! dictionary type
                # iterate over group contributions
                for group_name, group_value in group_x.items():
                    # check if group exists in  parameters
                    if group_name in group_id.values():
                        # add to count dictionary
                        res_ = ZabranskyRuzickaGroupData(
                            id=group_name,
                            name=group_name,
                            count=float(group_value),
                            data=self._get_group_x_data(
                                group_name=group_name,
                                group_mode='contribution' if group_name in self.group_id.values() else 'correction'
                            )
                        )
                        # append
                        valid_groups[group_name] = res_
            elif isinstance(
                group_x,
                ZabranskyRuzickaGroupContributions,
            ):
                # ! dataclass type
                # iterate over dataclass fields
                for field_name, field_info in ZabranskyRuzickaGroupContributions.model_fields.items():
                    # get field alias ( group name)
                    alias = field_info.alias

                    # check if field is None
                    if alias is None:
                        continue

                    # value stored in the model instance = use field_name
                    group_unit: GroupUnit = getattr(
                        group_x,
                        field_name
                    )

                    # check if group value is greater than 0
                    if group_unit.value > 0:
                        # check if group exists in  parameters
                        if alias in group_id.values():
                            # add to count dictionary
                            res_ = ZabranskyRuzickaGroupData(
                                id=alias,
                                name=field_name,
                                count=float(group_unit.value),
                                data=self._get_group_x_data(
                                    group_name=alias,
                                    group_mode='contribution'
                                )
                            )
                            # append
                            valid_groups[alias] = res_
            elif isinstance(
                group_x,
                ZabranskyRuzickaGroupContributionsCorrections,
            ):
                # ! dataclass type
                # iterate over dataclass fields
                for field_name, field_info in ZabranskyRuzickaGroupContributionsCorrections.model_fields.items():
                    # get field alias ( group name)
                    alias = field_info.alias

                    # check if field is None
                    if alias is None:
                        continue

                    # value stored in the model instance = use field_name
                    group_unit: GroupUnit = getattr(
                        group_x,
                        field_name
                    )

                    # check if group value is greater than 0
                    if group_unit.value > 0:
                        # check if group exists in  parameters
                        if alias in group_id.values():
                            # add to count dictionary
                            res_ = ZabranskyRuzickaGroupData(
                                id=alias,
                                name=field_name,
                                count=float(group_unit.value),
                                data=self._get_group_x_data(
                                    group_name=alias,
                                    group_mode='correction'
                                )
                            )
                            # append
                            valid_groups[alias] = res_
            else:
                logger.error("Invalid type for group contributions!")
                return valid_groups

            return valid_groups
        except Exception as e:
            raise Exception("Checking group contributions failed!, ", e)

    def _build_delta_c(
            self,
    ):
        '''
        Calculates delta C.

        Parameters
        ----------
        temperature : float
            Temperature in Kelvin.
        '''
        try:
            # SECTION: calculate sigma
            # initialize critical temperature
            sigma = {}

            # iterate over valid groups
            for group_name, group_info in self.valid_groups.items():
                # get group count
                group_count = group_info.count
                # get contribution data
                contribution_data = group_info.data
                # get contribution
                val_a_i = contribution_data.get('a_i', 0)
                val_b_i = contribution_data.get('b_i', 0)
                val_d_i = contribution_data.get('d_i', 0)

                # accumulate to sigma
                sigma[group_name] = {
                    'a_i': float(val_a_i),
                    'b_i': float(val_b_i),
                    'd_i': float(val_d_i),
                    'count': group_count
                }

            return sigma
        except Exception as e:
            raise Exception("Calculating critical temperature failed!, ", e)

    def _calc_sigma_delta_c(
            self,
            delta_c: Dict[str, Dict[str, float]],
            temperature: float
    ):
        '''
        Calculates sigma delta C.

        Parameters
        ----------
        delta_c : Dict[str, Dict[str, float]]
            Delta C contributions.
        temperature : float
            Temperature in Kelvin.
        '''
        try:
            # SECTION: calculate sigma
            # initialize critical temperature
            sigma = {}

            # iterate over valid groups
            for group_name, data in delta_c.items():
                # get group count
                group_count = data.get('count', 0)
                # get contribution
                val_a_i = data.get('a_i', 0)
                val_b_i = data.get('b_i', 0)
                val_d_i = data.get('d_i', 0)

                # calculate delta C for the group
                delta_c_i = (
                    float(val_a_i) +
                    float(val_b_i) * (temperature/100) +
                    float(val_d_i) * pow(temperature/100, 2)
                ) * group_count

                # accumulate to sigma
                sigma[group_name] = delta_c_i

            return sigma
        except Exception as e:
            raise Exception("Calculating critical temperature failed!, ", e)

    def _calc(
        self
    ) -> EstimatedProp:
        '''
        Calculates properties using  method.

        Returns
        -------
        EstimatedProp
            Calculated thermodynamic properties.
        '''
        try:
            # SECTION: create function
            def Cp_LIQ(T: float) -> float:
                '''
                Calculates liquid heat capacity at temperature T.

                Parameters
                ----------
                T : float
                    Temperature in Kelvin.

                Returns
                -------
                Cp_LIQ : float
                    Liquid heat capacity in J/mol·K.
                '''
                # create sigma delta C
                delta_c = self._build_delta_c()

                # calculate sigma delta C
                sigma_delta_c = self._calc_sigma_delta_c(delta_c, T)

                # sum over all groups
                Cp_LIQ_value = sum(sigma_delta_c.values())
                # times 8.314472
                Cp_LIQ_value *= 8.314472

                return Cp_LIQ_value

            # return
            return EstimatedProp(
                value=Cp_LIQ,
                unit="J/mol·K",
                symbol="Cp_LIQ"
            )
        except Exception as e:
            raise Exception("Calculating properties failed!, ", e)



from .info import (
    __author__,
    __description__,
    __email__,
    __license__,
    __version__,
)

from .settings import (
    JOBACK_DATA_FILE,
    JOBACK_TABLE_COLUMN_GROUP,
    ZABRANSKY_RUZICKA_DATA_FILE_1,
    ZABRANSKY_RUZICKA_DATA_FILE_2,
    ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP,
    ZABRANSKY_RUZICKA_REFERENCE
)

__all__ = [
    "__author__",
    "__description__",
    "__email__",
    "__license__",
    "__version__",
    "JOBACK_DATA_FILE",
    "JOBACK_TABLE_COLUMN_GROUP",
    "ZABRANSKY_RUZICKA_DATA_FILE_1",
    "ZABRANSKY_RUZICKA_DATA_FILE_2",
    "ZABRANSKY_RUZICKA_TABLE_COLUMN_GROUP",
]
